from itertools import product

import numpy as np
import pytest

from mrrt.mri.operators import MRI_Cartesian
from mrrt.utils import config, masker, embed
from mrrt.utils import fftn, fftnc, ifftn, ifftnc

all_xp = [np]
if config.have_cupy:
    import cupy

    if cupy.cuda.runtime.getDeviceCount() > 0:
        all_xp += [cupy]


def get_loc(xp):
    """Location arguments corresponding to numpy or CuPy case."""
    if xp is np:
        return dict(loc_in="cpu", loc_out="cpu")
    else:
        return dict(loc_in="gpu", loc_out="gpu")


def get_data(xp, shape=(128, 127)):
    rstate = xp.random.RandomState(1234)
    # keep one dimension odd to make sure shifts work correctly
    c = rstate.randn(*shape)
    return c


if config.have_pyfftw:
    preplan_pyfftw_vals = [True, False]
else:
    preplan_pyfftw_vals = [False]


# @dec.slow
@pytest.mark.parametrize(
    "xp, shift, nd_in, nd_out, order, real_dtype",
    product(
        all_xp,
        [True, False],
        [True, False],
        [True, False],
        ["C", "F"],
        [np.float32, np.float64, np.complex64, np.complex128],
    ),
)
def test_FFT(xp, shift, nd_in, nd_out, order, real_dtype):
    rtol = 1e-4
    atol = 1e-4
    cimg = get_data(xp).astype(real_dtype)
    cplx_type = xp.result_type(cimg.dtype, xp.complex64)

    FTop = MRI_Cartesian(
        cimg.shape,
        order=order,
        use_fft_shifts=shift,
        nd_input=nd_in,
        nd_output=nd_out,
        gpu_force_reinit=False,
        mask_kspace_on_gpu=(not shift),
        **get_loc(xp),
    )

    """
    test forward transform
    """
    tmp = FTop * cimg
    if nd_out:
        assert tmp.shape == cimg.shape
    else:
        assert tmp.ndim == 1
    if xp.isrealobj(cimg):
        assert tmp.real.dtype == cimg.dtype
    else:
        assert tmp.dtype == cimg.dtype

    tmp = tmp.reshape(cimg.shape, order=order)
    if shift:
        numpy_tmp = xp.fft.fftshift(xp.fft.fftn(xp.fft.ifftshift(cimg)))
    else:
        numpy_tmp = xp.fft.fftn(cimg)

    if numpy_tmp.dtype != cplx_type:
        numpy_tmp = numpy_tmp.astype(cplx_type)

    xp.testing.assert_allclose(tmp, numpy_tmp, rtol=rtol, atol=atol)

    """
    test adjoint transform
    """
    tmp2 = FTop.H * tmp
    if nd_in:
        assert tmp2.shape == cimg.shape
    else:
        assert tmp2.ndim == 1

    if xp.isrealobj(cimg):
        assert tmp2.real.dtype == cimg.dtype
    else:
        assert tmp2.dtype == cimg.dtype

    tmp2 = tmp2.reshape(cimg.shape, order=order)
    if shift:
        numpy_tmp2 = xp.fft.fftshift(xp.fft.ifftn(xp.fft.ifftshift(numpy_tmp)))
    else:
        numpy_tmp2 = xp.fft.ifftn(numpy_tmp)
    xp.testing.assert_allclose(tmp2, numpy_tmp2, rtol=rtol, atol=atol)


@pytest.mark.parametrize(
    "xp, shift, nd_in, nd_out, order",
    product(all_xp, [True, False], [True, False], [True, False], ["C", "F"]),
)
def test_FFT_roundtrips(xp, shift, nd_in, nd_out, order):
    rtol = 1e-3
    atol = 1e-3
    c = get_data(xp)

    FTop = MRI_Cartesian(
        c.shape,
        order=order,
        nd_input=nd_in,
        nd_output=nd_out,
        gpu_force_reinit=False,
        mask_kspace_on_gpu=False,
        **get_loc(xp),
    )

    """
    test round trip transform with paranthesis grouping
    """
    tmp = FTop.H * (FTop * c)
    xp.testing.assert_allclose(
        tmp.reshape(c.shape, order=order).real, c, rtol=rtol, atol=atol
    )

    """
    test round trip without paranthesis
    """
    tmp = FTop.H * FTop * c
    xp.testing.assert_allclose(
        tmp.reshape(c.shape, order=order).real, c, rtol=rtol, atol=atol
    )

    """
    test round trip with composite operator
    """
    tmp = (FTop.H * FTop) * c
    xp.testing.assert_allclose(
        tmp.reshape(c.shape, order=order).real, c, rtol=rtol, atol=atol
    )


# @dec.slow
@pytest.mark.parametrize(
    "xp, shift, nd_in, nd_out, order, ortho, fft_axes",
    product(
        all_xp,
        [True, False],
        [True, False],
        [True, False],
        ["C", "F"],
        [True, False],
        [(0,), (1,), (0, 1), None],
    ),
)
def test_FFT_axes_subsets_and_ortho(
    xp, shift, nd_in, nd_out, order, ortho, fft_axes
):
    rtol = 1e-3
    atol = 1e-3
    c = get_data(xp)
    """ Test applying FFT only along particular axes. """
    FTop = MRI_Cartesian(
        c.shape,
        order=order,
        use_fft_shifts=shift,
        nd_input=nd_in,
        nd_output=nd_out,
        fft_axes=fft_axes,
        ortho=ortho,
        gpu_force_reinit=False,
        mask_kspace_on_gpu=(not shift),
        **get_loc(xp),
    )

    """
    test forward transform
    """
    tmp = FTop * c
    if nd_out:
        assert tmp.shape == c.shape
    else:
        assert tmp.ndim == 1
    assert tmp.real.dtype == c.dtype

    tmp = tmp.reshape(c.shape, order=order)
    if ortho:
        fftargs = dict(norm="ortho")
    else:
        fftargs = {}
    if shift:
        numpy_tmp = xp.fft.fftshift(
            xp.fft.fftn(
                xp.fft.ifftshift(c, axes=fft_axes), axes=fft_axes, **fftargs
            ),
            axes=fft_axes,
        )
    else:
        numpy_tmp = xp.fft.fftn(c, axes=fft_axes, **fftargs)
    xp.testing.assert_allclose(tmp, numpy_tmp, rtol=rtol, atol=atol)

    """
    test adjoint transform
    """
    tmp2 = FTop.H * numpy_tmp
    if nd_in:
        assert tmp2.shape == c.shape
    else:
        assert tmp2.ndim == 1
    assert tmp2.real.dtype == c.dtype

    tmp2 = tmp2.reshape(c.shape, order=order)
    if shift:
        numpy_tmp2 = xp.fft.fftshift(
            xp.fft.ifftn(
                xp.fft.ifftshift(numpy_tmp, axes=fft_axes),
                axes=fft_axes,
                **fftargs,
            ),
            axes=fft_axes,
        )
    else:
        numpy_tmp2 = xp.fft.ifftn(numpy_tmp, axes=fft_axes, **fftargs)
    xp.testing.assert_allclose(tmp2, numpy_tmp2, rtol=rtol, atol=atol)


@pytest.mark.parametrize(
    "xp, shift, nd_in, nd_out, order",
    product(all_xp, [True, False], [True, False], [True, False], ["C", "F"]),
)
def test_FFT_2reps(xp, shift, nd_in, nd_out, order):
    """ multiple input case. """
    c = get_data(xp)
    FTop = MRI_Cartesian(
        c.shape,
        order=order,
        use_fft_shifts=shift,
        nd_input=nd_in,
        nd_output=nd_out,
        gpu_force_reinit=False,
        mask_kspace_on_gpu=(not shift),
        **get_loc(xp),
    )

    """
    test forward transform with 2 repetitions
    """
    nreps = 2
    if order == "F":
        c2 = np.stack((c,) * nreps, axis=-1)
    else:
        c2 = np.stack((c,) * nreps, axis=0)
    tmp = FTop * c2
    if nd_out:
        if order == "F":
            assert tmp.shape == c.shape + (nreps,)
        else:
            assert tmp.shape == (nreps,) + c.shape
    else:
        assert tmp.ndim == 2
        if order == "F":
            assert tmp.shape[-1] == nreps
        else:
            assert tmp.shape[0] == nreps
    """
    test adjoint transform with 2 repetitions
    """
    tmp2 = FTop.H * tmp
    if nd_in:
        if order == "F":
            assert tmp2.shape == c.shape + (nreps,)
        else:
            assert tmp2.shape == (nreps,) + c.shape
    else:
        assert tmp2.ndim == 2

    # scipy compatibility
    xp.testing.assert_array_equal(tmp, FTop.matvec(c2))
    xp.testing.assert_array_equal(tmp2, FTop.rmatvec(tmp))


# @dec.slow
@pytest.mark.parametrize(
    "xp, shift, nd_in, nd_out, loop, preplan_pyfftw, order",
    product(
        all_xp,
        [True, False],
        [True, False],
        [True, False],
        [True, False],
        preplan_pyfftw_vals,
        ["C", "F"],
    ),
)
def test_FFT_coilmap(xp, shift, nd_in, nd_out, loop, preplan_pyfftw, order):
    """ multiple coil  case. """
    ncoils = 4
    c = get_data(xp)
    if order == "F":
        cmap = 1 / ncoils * xp.ones((c.shape + (ncoils,)), dtype=c.dtype)
    else:
        cmap = 1 / ncoils * xp.ones(((ncoils,) + c.shape), dtype=c.dtype)

    if xp != np and preplan_pyfftw:
        # pyFFTW preplan case doesn't apply to the GPU
        return

    if order == "C":
        cmap = xp.ascontiguousarray(cmap)
    else:
        cmap = xp.asfortranarray(cmap)
    FTop = MRI_Cartesian(
        c.shape,
        order=order,
        use_fft_shifts=shift,
        nd_input=nd_in,
        nd_output=nd_out,
        loop_over_coils=loop,
        preplan_pyfftw=preplan_pyfftw,
        coil_sensitivities=cmap,
        gpu_force_reinit=False,
        disable_warnings=True,
        mask_kspace_on_gpu=(not shift),
        **get_loc(xp),
    )
    """
    test forward transform with 4 coils
    """
    tmp = FTop * c
    if nd_out:
        if order == "F":
            assert tmp.shape == c.shape + (ncoils,)
        else:
            assert tmp.shape == (ncoils,) + c.shape
    else:
        assert tmp.ndim == 1

    """
    test adjoint transform
    """
    tmp2 = FTop.H * tmp
    if nd_in:
        assert tmp2.shape == c.shape
    else:
        assert tmp2.ndim == 1
    assert tmp2.real.dtype == c.dtype


# @dec.slow
@pytest.mark.parametrize(
    "xp, shift, nd_in, nd_out, loop, preplan_pyfftw, order",
    product(
        all_xp,
        [True, False],
        [True, False],
        [True, False],
        [True, False],
        preplan_pyfftw_vals,
        ["C", "F"],
    ),
)
def test_FFT_coilmap_2reps(
    xp, shift, nd_in, nd_out, loop, preplan_pyfftw, order
):
    """ multiple coil, multiple input case. """
    ncoils = 4
    c = get_data(xp)
    if order == "F":
        cmap = 1 / ncoils * xp.ones((c.shape + (ncoils,)), dtype=c.dtype)
    else:
        cmap = 1 / ncoils * xp.ones(((ncoils,) + c.shape), dtype=c.dtype)

    if xp != np and preplan_pyfftw:
        # pyFFTW preplan case doesn't apply to the GPU
        return

    if order == "C":
        cmap = xp.ascontiguousarray(cmap)
    else:
        cmap = xp.asfortranarray(cmap)
    FTop = MRI_Cartesian(
        c.shape,
        order=order,
        use_fft_shifts=shift,
        nd_input=nd_in,
        nd_output=nd_out,
        loop_over_coils=loop,
        preplan_pyfftw=preplan_pyfftw,
        coil_sensitivities=cmap,
        gpu_force_reinit=False,
        disable_warnings=True,
        mask_kspace_on_gpu=(not shift),
        **get_loc(xp),
    )
    """
    test forward transform with 4 coils and 2 repetitions
    """
    nreps = 2

    if order == "F":
        c2 = np.stack((c,) * nreps, axis=-1)
    else:
        c2 = np.stack((c,) * nreps, axis=0)
    tmp = FTop * c2
    if order == "F":
        if nd_out:
            assert tmp.shape == c.shape + (ncoils, nreps)
        else:
            assert tmp.ndim == 2
            assert tmp.shape[-1] == nreps  # ncoils is folded into first dim
    else:
        if nd_out:
            assert tmp.shape == (nreps, ncoils) + c.shape
        else:
            assert tmp.ndim == 2
            assert tmp.shape[0] == nreps  # ncoils is folded into first dim
    """
    test adjoint transform with 2 repetitions
    """
    tmp2 = FTop.H * tmp
    if nd_in:
        if order == "F":
            assert tmp2.shape == c.shape + (nreps,)
        else:
            assert tmp2.shape == (nreps,) + c.shape
    else:
        assert tmp2.ndim == 2


@pytest.mark.parametrize(
    "xp, shift, nd_in, nd_out, order",
    product(all_xp, [True, False], [True, False], [True, False], ["C", "F"]),
)
def test_FFT_composite(xp, shift, nd_in, nd_out, order):
    """Testing composite forward and adjoint operator."""
    rtol = 1e-4
    atol = 1e-4
    c = get_data(xp)
    FTop = MRI_Cartesian(
        c.shape,
        order=order,
        use_fft_shifts=shift,
        nd_input=nd_in,
        nd_output=nd_out,
        gpu_force_reinit=False,
        mask_kspace_on_gpu=(not shift),
        **get_loc(xp),
    )

    # create new linear operator for forward followed by inverse transform
    FtF = FTop.H * FTop
    """
    test forward transform
    """
    tmp = FtF * c
    if nd_in:
        assert tmp.shape == c.shape
    else:
        assert tmp.ndim == 1
    assert tmp.real.dtype == c.dtype

    tmp = tmp.reshape(c.shape, order=order)
    xp.testing.assert_allclose(tmp, c, rtol=rtol, atol=atol)


@pytest.mark.parametrize(
    "xp, shift, nd_in, order",
    product(all_xp, [True, False], [True, False], ["C", "F"]),
)
def test_partial_FFT_allsamples(xp, shift, nd_in, order):
    rtol = 1e-4
    atol = 1e-4
    c = get_data(xp)
    sample_mask = xp.ones(c.shape)
    """ masked FFT without missing samples """
    # TODO: when all samples are present, can test against normal FFT
    nd_out = False
    FTop = MRI_Cartesian(
        c.shape,
        order=order,
        use_fft_shifts=shift,
        nd_input=nd_in,
        nd_output=nd_out,
        sample_mask=sample_mask,
        gpu_force_reinit=False,
        mask_kspace_on_gpu=(not shift),
        **get_loc(xp),
    )  # no missing samples

    # create new linear operator for forward followed by inverse transform
    FtF = FTop.H * FTop

    """
    test forward transform
    """
    tmp = FtF * c
    if nd_in:
        assert tmp.shape == c.shape
    else:
        assert tmp.ndim == 1
    assert tmp.real.dtype == c.dtype

    tmp = tmp.reshape(c.shape, order=order)
    xp.testing.assert_allclose(tmp, c, rtol=rtol, atol=atol)


@pytest.mark.parametrize(
    "xp, shift, nd_in, order",
    product(all_xp, [True, False], [True, False], ["C", "F"]),
)
def test_partial_FFT(xp, shift, nd_in, order):
    """ masked FFT with missing samples """
    # TODO: check accuracy against brute force DFT
    c = get_data(xp)
    rstate = xp.random.RandomState(1234)
    sample_mask = rstate.rand(*(128, 127)) > 0.5

    nd_out = False
    FTop = MRI_Cartesian(
        c.shape,
        order=order,
        use_fft_shifts=shift,
        nd_input=nd_in,
        nd_output=nd_out,
        sample_mask=sample_mask,
        gpu_force_reinit=False,
        mask_kspace_on_gpu=(not shift),
        **get_loc(xp),
    )

    # create new linear operator for forward followed by inverse transform
    FtF = FTop.H * FTop

    """
    test round trip transform
    """
    tmp = FtF * c
    if nd_in:
        assert tmp.shape == c.shape
    else:
        assert tmp.ndim == 1
    assert tmp.real.dtype == c.dtype

    tmp = tmp.reshape(c.shape, order=order)


@pytest.mark.parametrize(
    "xp, shift, nd_in, order",
    product(
        all_xp,
        [True, False],
        [True, False],
        ["F"],  # TODO: also implement order = 'C' case?
    ),
)
def test_partial_FFT_with_im_mask(xp, shift, nd_in, order):
    """ masked FFT with missing samples and masked image domain """
    # TODO: check accuracy against brute force DFT
    c = get_data(xp)
    rstate = xp.random.RandomState(1234)
    sample_mask = rstate.rand(*(128, 127)) > 0.5

    x, y = xp.meshgrid(
        xp.arange(-c.shape[0] // 2, c.shape[0] // 2),
        xp.arange(-c.shape[1] // 2, c.shape[1] // 2),
        indexing="ij",
        sparse=True,
    )

    # make a circular mask
    im_mask = xp.sqrt(x * x + y * y) < c.shape[0] // 2

    nd_out = False
    FTop = MRI_Cartesian(
        c.shape,
        order=order,
        im_mask=im_mask,
        use_fft_shifts=shift,
        nd_input=nd_in,
        nd_output=nd_out,
        sample_mask=sample_mask,
        gpu_force_reinit=False,
        mask_kspace_on_gpu=(not shift),
        **get_loc(xp),
    )

    # test forward only
    forw = embed(
        FTop * masker(c, im_mask, order=order), sample_mask, order=order
    )
    if shift:
        expected_forw = sample_mask * fftnc(c * im_mask)
    else:
        expected_forw = sample_mask * fftn(c * im_mask)
    xp.testing.assert_allclose(forw, expected_forw, rtol=1e-7, atol=1e-4)

    # test roundtrip
    roundtrip = FTop.H * (FTop * masker(c, im_mask, order=order))
    if shift:
        expected_roundtrip = masker(
            ifftnc(sample_mask * fftnc(c * im_mask)), im_mask, order=order
        )
    else:
        expected_roundtrip = masker(
            ifftn(sample_mask * fftn(c * im_mask)), im_mask, order=order
        )
    xp.testing.assert_allclose(
        roundtrip, expected_roundtrip, rtol=1e-7, atol=1e-4
    )

    # TODO: grlee77: fix FtF operation on masked data and uncomment case below
    # # create new linear operator for forward followed by inverse transform
    # FtF = FTop.H * FTop
    # roundtrip2 = FtF * masker(c, im_mask, order=order)
    # xp.testing.assert_allclose(
    #     roundtrip2, expected_roundtrip, rtol=1e-7, atol=1e-4
    # )

    # test roundtrip with 2 reps
    c2 = xp.stack([c] * 2, axis=-1)
    roundtrip = FTop.H * (FTop * masker(c2, im_mask, order=order))
    if shift:
        expected_roundtrip = masker(
            ifftnc(
                sample_mask[..., xp.newaxis]
                * fftnc(c2 * im_mask[..., xp.newaxis], axes=(0, 1)),
                axes=(0, 1),
            ),
            im_mask,
            order=order,
        )
    else:
        expected_roundtrip = masker(
            ifftn(
                sample_mask[..., xp.newaxis]
                * fftn(c2 * im_mask[..., xp.newaxis], axes=(0, 1)),
                axes=(0, 1),
            ),
            im_mask,
            order=order,
        )
    xp.testing.assert_allclose(
        roundtrip, expected_roundtrip, rtol=1e-7, atol=1e-4
    )


@pytest.mark.parametrize(
    "xp, shift, nd_in, order",
    product(all_xp, [True, False], [True, False], ["C", "F"]),
)
def test_partial_FFT_coils(xp, shift, nd_in, order):
    """ masked FFT with missing samples and multiple coils """
    # TODO: check accuracy against brute force DFT
    ncoils = 4
    c = get_data(xp)
    if order == "F":
        cmap = 1 / ncoils * xp.ones((c.shape + (ncoils,)), dtype=c.dtype)
    else:
        cmap = 1 / ncoils * xp.ones(((ncoils,) + c.shape), dtype=c.dtype)
    rstate = xp.random.RandomState(1234)
    sample_mask = rstate.rand(*(128, 127)) > 0.5

    nd_out = False
    if order == "C":
        cmap = xp.ascontiguousarray(cmap)
    else:
        cmap = xp.asfortranarray(cmap)
    FTop = MRI_Cartesian(
        c.shape,
        order=order,
        use_fft_shifts=shift,
        nd_input=nd_in,
        nd_output=nd_out,
        sample_mask=sample_mask,
        coil_sensitivities=cmap,
        gpu_force_reinit=False,
        disable_warnings=True,
        mask_kspace_on_gpu=(not shift),
        **get_loc(xp),
    )

    # create new linear operator for forward followed by inverse transform
    FtF = FTop.H * FTop

    """
    test round-trip transform
    """
    tmp = FtF * c
    if nd_in:
        assert tmp.shape == c.shape
    else:
        assert tmp.ndim == 1

    y = FTop * c

    assert tmp.real.dtype == c.dtype

    # now test with two sets of coil maps (e.g. ESPIRIT Soft-SENSE)
    for M in [2, 3]:
        if order == "F":
            cmap_multi = xp.asfortranarray(xp.stack((cmap,) * M, axis=-1))
            c_multi = xp.asfortranarray(xp.stack((c,) * M, axis=-1))
        else:
            cmap_multi = xp.asfortranarray(xp.stack((cmap,) * M, axis=0))
            c_multi = xp.asfortranarray(xp.stack((c,) * M, axis=0))

        FTop_multi = MRI_Cartesian(
            c.shape,
            order=order,
            use_fft_shifts=shift,
            nd_input=nd_in,
            nd_output=nd_out,
            sample_mask=sample_mask,
            coil_sensitivities=cmap_multi,
            gpu_force_reinit=False,
            disable_warnings=True,
            mask_kspace_on_gpu=(not shift),
            **get_loc(xp),
        )
        y_multi = FTop_multi * c_multi

        # identical data and cmap for each set, so y_multi should just be
        # M * y
        assert xp.max(xp.abs(y_multi - M * y)) < 1e-4

        FtF = FTop_multi.H * FTop_multi
        tmp = FtF * c_multi
        if nd_in:
            assert tmp.shape == c_multi.shape
        else:
            assert tmp.ndim == 1



import numpy as np
from mrrt.utils import prod

# TODO: remove unused sinc case


class PixelBasis:
    r""" Pixel basis for representing a continuous-to-discrete mapping between
    a continuous image and its sampled (discrete) values.
    """

    def __init__(self, kspace, pixel_basis="dirac", dx=None, fov=None, Nd=None):
        """ Initialize a pixel-basis and its transform.

        Parameters
        ----------
        kspace : array
            2D array of k-space coordinates :math:`[n_{samples}, n_{dim}]`.
        pixel_basis : {'dirac', 'rect'}
            basis type
        dx : array_like, optional
            image pixel spacings
        fov : array_like, optional
            image field of view (i.e. `Nd` * `dx`)
        Nd : array_like, optional
            image matrix size along each dimension

        Attributes
        ----------
        type : str
            basis type
        dx : array or None
            pixel spacing in image domain
        fov : array or None
            image field of view (i.e. `Nd` * `dx`)
        Nd : array or None
            image matrix size along each dimension
        transform : array
            Fourier transform of the basis evaluated at the k-space sample
            locations.

        Notes
        -----
        A continuous function, :math:`f(\vec{x})`, is represented by a
        discrete set of pixel values :math:`x_{i}` in basis,
        :math:`b(\vec{x})`

        .. math::

            f\left(\vec{x}\right)=\sum_{i=1}^{N}\,x_{i}b\left(\vec{x}-x_{i}\right)

        units for `dx`, `fov` must match (e.g. mm).
        `kspace` units (e.g. mm^-1) should be the inverse of those for `dx`,
        `fov`.

        References
        ----------
        .. [1] Pruessmann KP, Weiger M, Scheidegger MB, Boesiger P.  SENSE:
           Sensitivity Encoding for Fast MRI.  Magn. Reson. Med. 1999;
           42:952-962.
        .. [2] Sutton, BP.  Physics Based Iterative Reconstruction for MRI:
           Compensating and Estimating Field Inhomogeneity and T2\* Relaxation.
           Doctoral Dissertation.  University of Michigan. 2003.
        .. [3] Barrett HH, Myers KJ.  "7.1: Objects and Images" In *Foundations
           of Image Science*.  2003. Wiley-Interscience.

        """
        self.type = pixel_basis
        self.dx = dx
        self.fov = fov
        self.Nd = Nd
        if self.dx is not None:
            self.dx = np.atleast_1d(self.dx)
        if self.fov is not None:
            self.fov = np.atleast_1d(self.fov)
        if self.Nd is not None:
            self.Nd = np.atleast_1d(self.Nd)
        else:
            self.Nd = np.round(self.fov / self.dx).astype(int)
        self.kspace = kspace
        if kspace.ndim != 2:
            raise ValueError("kspace must be 2D: [nsamples x ndim]")

        nk, ndim = kspace.shape
        if self.type == "dirac":
            self.transform = None
        elif self.type == "rect":
            if self.dx is None:
                if (self.fov is None) or (self.Nd is None):
                    raise ValueError("must specify dx or both fov and Nd")
                self.dx = self.fov / self.Nd
            if len(self.dx) != ndim:
                if len(self.dx) == 1:
                    self.dx = np.asarray(self.dx.tolist() * ndim)
                else:
                    raise ValueError("len(dx) != ndim")
            self.transform = np.ones((nk,))

            for idx in range(ndim):
                # allow for zero-sized pixels (Dirac impulses)
                if self.dx[idx] > 0:
                    # dx* sinc(dx*kx) term in Eq. 6.15 of Brad Sutton's thesis
                    self.transform = self.transform * (
                        self.dx[idx] * np.sinc(self.dx[idx] * kspace[:, idx])
                    )

            # make the average scaling close to 1
            self.transform /= self.transform.mean()
        elif self.type in ["sinc", "dirac*dx"]:
            # simply provide an appropriate "scale factor" dx to relate Fourier
            # integral and summation
            self.transform = np.ones((nk,)) * prod(
                f / n for f, n in zip(self.fov, self.Nd)
            )

            # make the average scaling close to 1
            self.transform /= self.transform.mean()
        else:
            raise ValueError("unknown PixelBasis type: {}".format(pixel_basis))

    def __str__(self):
        repstr = "Pixel Basis:\ntype={}, ndim={}, dx={}".format(
            self.type, len(self.Nd), self.dx
        )
        return repstr

    def __repr__(self):
        # repstr = ("PixelBasis(kspace, pixel_basis={}, ".format(self.type) +
        #           "dx={}, fov={}, Nd={})".format(self.fov, self.dx, self.Nd))
        repstr = (
            "PixelBasis(kspace, pixel_basis='{}', dx={}, fov={}, Nd={})"
        ).format(self.type, self.dx, self.fov, self.Nd)

        return repstr



"""Tests related to Non-Cartesian MRI reconstruction."""
from itertools import product
import time

import numpy as np
from numpy.testing import assert_, assert_equal
import pytest

from mrrt.operators.LinOp import DiagonalOperatorMulti
from mrrt.mri.operators.tests._generate_testdata import generate_sim_data
from mrrt.nufft import dtft, dtft_adj
from mrrt.utils import embed, have_cupy, profile

import os

OMIT_CPU = int(os.environ.get("OMIT_CPU", False))
OMIT_GPU = int(os.environ.get("OMIT_GPU", False))

all_xp = [np]
cpu_cases = ["CPU,Tab0", "CPU,Tab", "CPU,Sp"] if not OMIT_CPU else []
all_cases = cpu_cases
if have_cupy:
    import cupy

    if cupy.cuda.runtime.getDeviceCount() > 0 and not OMIT_GPU:
        gpu_cases = ["GPU,Tab0", "GPU,Tab", "GPU,Sp"]
        all_cases += gpu_cases
        all_xp += [cupy]

# To ignore PendingDeprecationWarning related to scipy.sparse we use:
#     @pytest.mark.filterwarnings("ignore:the matrix subclass is not")
#
# This class of warnings could also be ignored via the command line, e.g.:
#     pytest -W ignore::PendingDeprecationWarning test_mri_reconstruction.py


@profile
def _test_mri_multi(
    ndim=3,
    N0=8,
    grid_os_factor=1.5,
    J0=4,
    Ld=4096,
    n_coils=1,
    fieldmap_segments=None,
    precisions=["single", "double"],
    phasings=["real", "complex"],
    recon_cases=["CPU,Tab0", "CPU,Tab", "CPU,Sp"],
    rtol=1e-3,
    compare_to_exact=False,
    show_figures=False,
    nufft_kwargs={},
    navg_time=1,
    n_creation=1,
    return_errors=False,
    gpu_memflags=None,
    verbose=False,
    return_operator=False,
    spectral_offsets=None,
):
    """Run a batch of NUFFT tests."""
    all_err_forward = np.zeros(
        (len(recon_cases), len(precisions), len(phasings))
    )
    all_err_adj = np.zeros((len(recon_cases), len(precisions), len(phasings)))
    alltimes = {}
    if not np.isscalar(navg_time):
        navg_time_cpu, navg_time_gpu = navg_time
    else:
        navg_time_cpu = navg_time_gpu = navg_time
    for i, recon_case in enumerate(recon_cases):
        if "CPU" in recon_case:
            navg_time = navg_time_cpu
        else:
            navg_time = navg_time_gpu

        for j, precision in enumerate(precisions):
            for k, phasing in enumerate(phasings):
                if verbose:
                    print(
                        "phasing={}, precision={}, type={}".format(
                            phasing, precision, recon_case
                        )
                    )

                if "Tab" in recon_case:
                    # may want to create twice when benchmarking GPU case
                    # because the custom kernels are compiled the first time
                    ncr_max = n_creation
                else:
                    ncr_max = 1
                # on_gpu = ('GPU' in recon_case)
                for ncr in range(ncr_max):
                    (
                        Gn,
                        wi_full,
                        xTrue,
                        ig,
                        data_true,
                        times,
                    ) = generate_sim_data(
                        recon_case=recon_case,
                        ndim=ndim,
                        N0=N0,
                        J0=J0,
                        grid_os_factor=grid_os_factor,
                        fieldmap_segments=fieldmap_segments,
                        Ld=Ld,
                        n_coils=n_coils,
                        precision=precision,
                        phasing=phasing,
                        nufft_kwargs=nufft_kwargs,
                        MRI_object_kwargs=dict(gpu_memflags=gpu_memflags),
                        spectral_offsets=spectral_offsets,
                    )

                xp = Gn.xp

                # time the forward operator
                sim_data = Gn * xTrue  # dry run
                tstart = time.time()
                for nt in range(navg_time):
                    sim_data = Gn * xTrue
                    sim_data += 0.0
                sim_data = xp.squeeze(sim_data)  # TODO: should be 1D already?
                # print("type(xTrue) = {}".format(type(xTrue)))
                # print("type(sim_data) = {}".format(type(sim_data)))
                t_for = (time.time() - tstart) / navg_time
                times["MRI: forward"] = t_for

                # time the norm operator
                Gn.norm(xTrue)  # dry run
                tstart = time.time()
                for nt in range(navg_time):
                    Gn.norm(xTrue)
                t_norm = (time.time() - tstart) / navg_time

                times["MRI: norm"] = t_norm
                if precision == "single":
                    dtype_real = np.float32
                    dtype_cplx = np.complex64
                else:
                    dtype_real = np.float64
                    dtype_cplx = np.complex128

                if "Tab" in recon_case:
                    if phasing == "complex":
                        assert_equal(Gn.Gnufft.h[0].dtype, dtype_cplx)
                    else:
                        assert_equal(Gn.Gnufft.h[0].dtype, dtype_real)
                else:
                    if phasing == "complex":
                        assert_equal(Gn.Gnufft.p.dtype, dtype_cplx)
                    else:
                        assert_equal(Gn.Gnufft.p.dtype, dtype_real)
                assert_equal(sim_data.dtype, dtype_cplx)

                if compare_to_exact:
                    # compare_to_exact only currently for single-coil,
                    # no fieldmap case
                    if spectral_offsets is not None:
                        raise NotImplementedError(
                            "compare_to_exact doesn't currently support "
                            "spectral offsets"
                        )
                    nshift_exact = tuple(s / 2 for s in Gn.Nd)
                    sim_data2 = dtft(
                        xTrue, Gn.omega, shape=Gn.Nd, n_shift=nshift_exact
                    )

                    sd2_norm = xp.linalg.norm(sim_data2)
                    rel_err = xp.linalg.norm(sim_data - sim_data2) / sd2_norm
                    if "GPU" in recon_case:
                        if hasattr(rel_err, "get"):
                            rel_err = rel_err.get()
                    all_err_forward[i, j, k] = rel_err
                    print(
                        "{},{},{}: forward error = {}".format(
                            recon_case, precision, phasing, rel_err
                        )
                    )
                    rel_err_mag = (
                        xp.linalg.norm(np.abs(sim_data) - np.abs(sim_data2))
                        / sd2_norm
                    )
                    print(
                        f"{recon_case},{precision},{phasing}: "
                        f"forward mag diff error = {rel_err_mag}"
                    )
                    assert rel_err < rtol

                # TODO: update DiagonalOperatorMulti to auto-set loc_in,
                #       loc_out appropriately
                if xp is np:
                    diag_args = dict(loc_in="cpu", loc_out="cpu")
                else:
                    diag_args = dict(loc_in="gpu", loc_out="gpu")
                diag_op = DiagonalOperatorMulti(wi_full, **diag_args)
                if n_coils == 1:
                    data_dcf = diag_op * data_true
                else:
                    data_dcf = diag_op * sim_data

                # time the adjoint operation
                im_est = Gn.H * data_dcf  # dry run
                tstart = time.time()
                for nt in range(navg_time):
                    im_est = Gn.H * data_dcf
                t_adj = (time.time() - tstart) / navg_time
                times["MRI: adjoint"] = t_adj

                if hasattr(Gn, "mask") and Gn.mask is not None:
                    im_est = embed(im_est, Gn.mask)
                else:
                    if spectral_offsets is None:
                        im_est = im_est.reshape(Gn.Nd, order=Gn.order)
                    else:
                        im_est = im_est.reshape(
                            tuple(Gn.Nd) + (len(spectral_offsets),),
                            order=Gn.order,
                        )

                if compare_to_exact:
                    im_est_exact = dtft_adj(
                        data_dcf, Gn.omega, shape=Gn.Nd, n_shift=nshift_exact
                    )
                    ex_norm = xp.linalg.norm(im_est_exact)
                    rel_err = xp.linalg.norm(im_est - im_est_exact) / ex_norm
                    all_err_adj[i, j, k] = rel_err
                    if verbose:
                        print(
                            "{},{},{}: adjoint error = {}".format(
                                recon_case, precision, phasing, rel_err
                            )
                        )
                    rel_err_mag = (
                        xp.linalg.norm(np.abs(im_est) - np.abs(im_est_exact))
                        / ex_norm
                    )
                    if verbose:
                        print(
                            "{},{},{}: adjoint mag diff error = {}".format(
                                recon_case, precision, phasing, rel_err
                            )
                        )
                    assert_(rel_err < rtol)

                title = ", ".join([recon_case, precision, phasing])
                if show_figures:
                    from matplotlib import pyplot as plt
                    from pyvolplot import volshow

                    if compare_to_exact:
                        volshow(
                            [
                                im_est_exact,
                                im_est,
                                im_est_exact - im_est,
                                xp.abs(im_est_exact) - xp.abs(im_est),
                            ]
                        )
                    else:
                        volshow(im_est)
                        plt.title(title)
                alltimes[title] = times

    if return_operator:
        if return_errors:
            return Gn, alltimes, all_err_forward, all_err_adj
        return Gn, alltimes

    if return_errors:
        return alltimes, all_err_forward, all_err_adj
    return alltimes


@pytest.mark.parametrize(
    "recon_case, precision, phasing",
    product(all_cases, ["single", "double"], ["complex", "real"]),
)
@pytest.mark.filterwarnings("ignore:the matrix subclass is not")
def test_mri_2d_nocoils_nofieldmap_nocompare(
    recon_case, precision, phasing, show_figures=False, verbose=False
):

    _test_mri_multi(
        ndim=2,
        N0=16,
        grid_os_factor=1.5,
        J0=6,
        Ld=4096,
        n_coils=1,
        fieldmap_segments=None,
        precisions=[precision],
        phasings=[phasing],
        recon_cases=[recon_case],
        show_figures=show_figures,
        verbose=verbose,
        compare_to_exact=False,
        nufft_kwargs={},
        rtol=1e-4,
    )


# @dec.slow
@pytest.mark.parametrize(
    "recon_case, precision, phasing",
    product(all_cases, ["single", "double"], ["complex", "real"]),
)
@pytest.mark.filterwarnings("ignore:the matrix subclass is not")
def test_mri_2d_nocoils_nofieldmap(
    recon_case, precision, phasing, show_figures=False, verbose=False
):

    _test_mri_multi(
        ndim=2,
        N0=16,
        grid_os_factor=1.5,
        J0=6,
        Ld=4096,
        n_coils=1,
        fieldmap_segments=None,
        precisions=[precision],
        phasings=[phasing],
        recon_cases=[recon_case],
        show_figures=show_figures,
        verbose=verbose,
        compare_to_exact=True,
        nufft_kwargs={},
        rtol=1e-3,
    )


# @dec.slow
@pytest.mark.parametrize(
    "recon_case, precision, phasing",
    product(all_cases, ["single", "double"], ["complex", "real"]),
)
@pytest.mark.filterwarnings("ignore:the matrix subclass is not")
def test_mri_2d_nocoils_nofieldmap_kernels(
    recon_case, precision, phasing, show_figures=False, verbose=False
):
    N = 32
    grid_os_factor = 2
    J = 6
    t, ef, ea = _test_mri_multi(
        ndim=2,
        N0=N,
        grid_os_factor=grid_os_factor,
        J0=J,
        Ld=512,
        n_coils=1,
        fieldmap_segments=None,
        precisions=[precision],
        phasings=[phasing],
        recon_cases=[recon_case],
        show_figures=show_figures,
        verbose=verbose,
        compare_to_exact=True,
        nufft_kwargs={},
        rtol=1e-2,
        return_errors=True,
    )


# @dec.slow
@pytest.mark.parametrize(
    "recon_case, precision, phasing",
    product(all_cases, ["single", "double"], ["complex", "real"]),
)
@pytest.mark.filterwarnings("ignore:the matrix subclass is not")
def test_mri_2d_multicoils_nofieldmap(
    recon_case, precision, phasing, show_figures=False, verbose=False
):
    _test_mri_multi(
        ndim=2,
        N0=16,
        grid_os_factor=1.5,
        J0=6,
        Ld=4096,
        n_coils=4,
        fieldmap_segments=None,
        precisions=[precision],
        phasings=[phasing],
        recon_cases=[recon_case],
        compare_to_exact=False,
        rtol=1e-3,
        show_figures=show_figures,
        verbose=verbose,
    )


# @dec.slow
@pytest.mark.parametrize(
    "recon_case, precision, phasing",
    product(all_cases, ["single", "double"], ["complex", "real"]),
)
@pytest.mark.filterwarnings("ignore:the matrix subclass is not")
def test_mri_2d_multicoils_fieldmap(
    recon_case, precision, phasing, show_figures=False, verbose=False
):
    _test_mri_multi(
        ndim=2,
        N0=16,
        grid_os_factor=1.5,
        J0=6,
        Ld=4096,
        n_coils=4,
        fieldmap_segments=6,
        precisions=[precision],
        phasings=[phasing],
        recon_cases=[recon_case],
        compare_to_exact=False,
        show_figures=show_figures,
        verbose=verbose,
    )


# @dec.slow
@pytest.mark.parametrize(
    "recon_case, precision, phasing",
    product(all_cases, ["single", "double"], ["complex", "real"]),
)
@pytest.mark.filterwarnings("ignore:the matrix subclass is not")
def test_mri_3d_nocoils_nofieldmap(
    recon_case, precision, phasing, show_figures=False, verbose=False
):
    if "Tab0" in recon_case:
        rtol = 1e-2
    else:
        rtol = 1e-3
    _test_mri_multi(
        ndim=3,
        N0=12,
        grid_os_factor=1.5,
        J0=4,
        Ld=4096,
        n_coils=1,
        fieldmap_segments=None,
        precisions=[precision],
        phasings=[phasing],
        recon_cases=[recon_case],
        compare_to_exact=True,
        show_figures=show_figures,
        rtol=rtol,
        verbose=verbose,
    )


# @dec.slow
@pytest.mark.parametrize(
    "recon_case, precision, phasing",
    product(all_cases, ["single", "double"], ["complex", "real"]),
)
@pytest.mark.filterwarnings("ignore:the matrix subclass is not")
def test_mri_3d_multicoils_nofieldmap(
    recon_case, precision, phasing, show_figures=False, verbose=False
):
    _test_mri_multi(
        ndim=3,
        N0=12,
        grid_os_factor=1.5,
        J0=4,
        Ld=4096,
        n_coils=4,
        fieldmap_segments=None,
        precisions=[precision],
        phasings=[phasing],
        recon_cases=[recon_case],
        compare_to_exact=False,
        show_figures=show_figures,
        verbose=verbose,
    )


# @dec.slow
@pytest.mark.parametrize(
    "recon_case, precision, phasing",
    product(all_cases, ["single", "double"], ["complex", "real"]),
)
@pytest.mark.filterwarnings("ignore:the matrix subclass is not")
def test_mri_3d_multicoils_fieldmap(
    recon_case, precision, phasing, show_figures=False, verbose=False
):
    _test_mri_multi(
        ndim=3,
        N0=12,
        grid_os_factor=1.5,
        J0=4,
        Ld=4096,
        n_coils=4,
        fieldmap_segments=6,
        precisions=[precision],
        phasings=[phasing],
        recon_cases=[recon_case],
        compare_to_exact=False,
        show_figures=show_figures,
        verbose=verbose,
    )



import warnings
import numpy as np

from mrrt.nufft import NufftBase
from mrrt.operators import LinearOperatorMulti
from mrrt.utils import fftn, ifftn, next_fast_len
from mrrt.utils import complexify, embed, prod, profile


def compute_Q(G, wi=None, Nd_os=2, Kd_os=1.35, J=5, **extra_nufft_kwargs):
    """Compute Q such that IFFT(Q*FFT(x)) = (G.H * G * x).

    Notes
    -----
    requires that G.Kd ~= 2*G.Nd for good accuracy.
    can get away with Kd_os < substantially less than 2

    References
    ----------
    ..[1] Wajer FTAW, Pruessmann KP. Major Speedup of Reconstruction for
    Sensitivity Encoding with Arbitrary Trajectories.
    Proc. Intl. Soc. Mag. Reson. Med. 9 (2001), p.767.

    ..[2] Eggers H, Boernert P, Boesiger P.  Comparison of Gridding- and
    Convolution-Based Iterative Reconstruction Algorithms For
    Sensitivity-Encoded Non-Cartesian Acquisitions.
    Proc. Intl. Soc. Mag. Reson. Med. 10 (2002)

    ..[3] Liu C, Moseley ME, Bammer R.  Fast SENSE Reconstruction Using Linear
    System Transfer Function.
    Proc. Intl. Soc. Mag. Reson. Med. 13 (2005), p.689.
    """
    from mrrt.mri.operators import MRI_Operator, NUFFT_Operator

    if isinstance(G, NUFFT_Operator):
        Gnufft_op = G
    elif isinstance(G, MRI_Operator):
        Gnufft_op = G.Gnufft
    else:
        raise ValueError("G must be an NUFFT_Operator or MRI_Operator")

    # need reasonably accurate gridding onto a 2x oversampled grid
    Nd = (Nd_os * Gnufft_op.Nd).astype(np.intp)
    Kd = next_fast_len((Kd_os * Nd).astype(np.intp))
    if any(k < 2 * n for (k, n) in zip(G.Kd, G.Nd)):
        warnings.warn(
            "Q operator unlikely to be accurate.  Recommend using G "
            "with a grid oversampling factor of 2"
        )
    if Nd_os != 2:
        warnings.warn("recommend keeping Nd_os=2")

    G2 = NUFFT_Operator(
        omega=Gnufft_op.om,
        Nd=Nd,
        Kd=Kd,
        Jd=(J,) * len(Nd),
        Ld=Gnufft_op.Ld,
        n_shift=Nd / 2,
        mode=Gnufft_op.mode,
        phasing="real",  # ONLY WORKS IF THIS IS REAL!
        **extra_nufft_kwargs,
    )

    if wi is None:
        wi = np.ones(Gnufft_op.om.shape[0], dtype=Gnufft_op._cplx_dtype)

    psft = G2.H * wi
    # TODO: allow DiagonalOperator too for weights
    psft = np.fft.fftshift(psft.reshape(G2.Nd, order=G2.order))
    return fftn(psft)


def compute_Q_v2(G, copy_X=True):
    """Alternative version of compute_Q.

    experimental:  not recommended over compute_Q()
    """
    from mrrt.nufft._nufft import nufft_adj

    ones = np.ones(G.kspace.shape[0], G.Gnufft._cplx_dtype)
    sf = np.sqrt(prod(G.Gnufft.Kd))
    return sf * fftn(nufft_adj(G.Gnufft, ones, copy_X=True, return_psf=True))


class NUFFT_Operator(NufftBase, LinearOperatorMulti):

    """ Linear Operator wrapper around NufftBase.

    Parameters
    ----------
    Nd : array_like
        image domain dimensions
    omega : array_like
        k-space coordinates normalized to the range [-pi, pi]
        e.g. if kspace is in mm^-1 and dx in mm:
        ``omega[:, d] = 2 * np.pi * kspace[:, d] * dx[d]``
    mask : array_like, optional
        logical support mask over the image
    kwargs : dict, optional
        any additional keyword arguments to pass onto `NufftBase`

    Attributes
    ----------
    mask
    dim

    """

    @profile
    def __init__(
        self,
        Nd,
        omega,
        mask=None,
        squeeze_reps=True,
        loc_in="cpu",
        loc_out="cpu",
        order="F",
        **kwargs,
    ):

        if loc_in == "cpu":
            xp = np
        else:
            import cupy

            xp = cupy

        # masked boolean if mask is True everywhere or no mask is provided
        if mask is not None:
            mask = xp.asarray(mask, dtype=bool)

        self.masked = (
            (mask is not None)
            and (not isinstance(mask, tuple))
            and (mask.size != xp.count_nonzero(mask))
        )
        if isinstance(mask, tuple):
            self.__mask = None
        else:
            self.__mask = mask
        self.__Nd = Nd

        nargout = omega.shape[0]
        if self.masked:
            nargin = xp.count_nonzero(mask)
        else:
            nargin = prod(self.__Nd)

        if order != "F":
            raise ValueError("NUFFT_Operator only supports order='F'.")

        # initialize NUFFT
        NufftBase.__init__(self, Nd=Nd, omega=omega, order=order, **kwargs)

        # set appropriate operations as initialized by NufftBase
        self.__matvec = self.fft
        self.__matvec_transp = self.adj
        self.__matvec_adj = self.adj

        # construct the linear operator
        LinearOperatorMulti.__init__(
            self,
            nargin,
            nargout,
            symmetric=False,
            hermetian=False,
            matvec=self.__matvec,
            matvec_adj=self.__matvec_adj,
            matvec_transp=self.__matvec_adj,
            nd_input=False,  # True,
            nd_output=False,
            loc_in=loc_in,
            loc_out=loc_out,
            shape_in=self.Nd,
            shape_out=(nargout, 1),
            order=order,
            matvec_allows_repetitions=True,
            # MRI_Operator code assumes squeeze_reps_* are False
            squeeze_reps=squeeze_reps,
            mask_in=self.mask,
            mask_out=None,
            dtype=self._cplx_dtype,
        )

    @property
    def mask(self, create=False):
        "The image domain mask corresponding to the operator."
        if self.__mask is None and create:
            return self.xp.ones(self.Nd, dtype=np.bool)
        else:
            return self.__mask

    @mask.setter
    def mask(self, mask):
        "The image domain mask corresponding to the operator."
        if mask is None:
            self.masked = False
        elif self.__mask is not None and mask.shape != self.__mask.shape:
            raise ValueError("Shape of mask inconsistent with existing object")
        elif mask.sum() == mask.size:
            # True everywhere, so don't need the mask
            self.masked = False
            mask = None
        else:
            self.masked = True
        self.__mask = mask

    @property
    def dim(self):
        "The shape corresponding to the operator."
        return self.shape

    def prep_toeplitz(self, weights=None, Kd_os=1.35, J=5):
        if weights is not None:
            weights = complexify(weights)
        self.Q = compute_Q(self, weights=weights, Kd_os=Kd_os, J=J)

    @profile
    def norm(self, x):
        # if not hasattr(self, 'Q') or self.Q is None:
        #     warnings.warn("Toeplitz Q did not exist, creating it...")
        #     self.prep_toeplitz()

        x = complexify(x)
        if self.masked:
            x = embed(x, self.mask, order=self.order)
        try:
            if x.size % self.nargin != 0:
                raise ValueError("wrong size input")
            Nrepetitions = x.size // self.nargin
        except IndexError:
            Nrepetitions = 1

        if hasattr(self, "Q"):
            slices = [slice(None)] * x.ndim
            for d in range(len(self.Nd)):
                slices[d] = slice(self.Nd[d])
            if Nrepetitions == 1:
                y = fftn(x, s=self.Q.shape)
                y *= self.Q
                y = ifftn(y)[slices]
            else:
                if self.order == "C":
                    x_shape = (Nrepetitions,) + tuple(self.Nd)
                else:
                    x_shape = tuple(self.Nd) + (Nrepetitions,)
                x = x.reshape(x_shape, order=self.order)
                fft_axes = tuple(np.arange(len(self.Nd)))
                y = fftn(x, s=self.Q.shape, axes=fft_axes)
                y *= self.Q[..., np.newaxis]  # add an axis for repetitions
                y = ifftn(y, axes=fft_axes)[slices]
        else:
            y = self.H * (self * x)
        return y



"""Non-Cartesian MRI Operator.

Non-Cartesian MRI Operator
==========================
Class implementing forward and adjoint operators for Non-Cartesian MRI. This
includes both coil maps and field maps for correction of off-resonance (and
potentially relaxation).

TODO: expand documentation here


"""

from __future__ import division, print_function, absolute_import
from math import ceil
import warnings
import numpy as np
from mrrt.mri import mri_exp_approx
from mrrt.mri.operators import NUFFT_Operator
from mrrt.utils import (
    complexify,
    get_array_module,
    have_cupy,
    masker,
    next_fast_len_multiple,
    power_method,
    prod,
    profile,
)

from mrrt.operators import BlockDiagLinOp, DiagonalOperator, LinearOperatorMulti
from mrrt.mri.operators._pixel_basis import PixelBasis

# from mrrt.nufft import compute_Q

if have_cupy:
    import cupy


# TODO: implement Toeplitz approach?

# TODO: allow fmap_basis to be only a single shot for multi-shot trajectories
# where each shot has the same temporal readout points.  This can be used to
# reduce GPU memory usage of fieldmap-corrected MRI

# TODO: There is an occasional non-reproducable failure on GPU only
#       for some table-based NUFFTs:
#       OMIT_CPU=1 py.test test_MRI_reconstruction.py -k "test_mri_2d_nocoils_nofieldmap_kernels" -v


def exp_xform(x, u, v, xp=None):
    xp, on_gpu = get_array_module(x, xp)

    x = xp.asarray(x)
    u = xp.asarray(u)
    v = xp.asarray(v)

    if x.ndim == 1:
        x = x[:, np.newaxis]

    if x.shape[0] != u.shape[1]:
        print("x.shape[0]=%d, u.shape[1]=%d." % (x.shape[0], u.shape[1]))
        raise ValueError("dimensions inconsistent")

    if u.shape[0] != v.shape[0]:
        print("u.shape[0]=%d, x.shape[0]=%d." % (u.shape[0], x.shape[0]))
        raise ValueError("dimensions inconsistent")

    tmp = -xp.dot(v.T, u)
    xp.exp(tmp, out=tmp)
    return xp.dot(tmp, x)


# @profile
def determine_fieldmap_segments(
    ti,
    zmap,
    mask=None,
    approx_type="default",
    fieldmap_segments_start=5,
    err_thresh=1e-3,
    exp_approx_args={},
    xp=np,
):
    """Determine an appropriate number of fieldmap segments."""
    fieldmap_segments = fieldmap_segments_start
    fieldmap_segments_adjust = 0
    error_acceptable = False
    if mask is not None:
        zmap = masker(zmap, mask, xp=xp)
    while error_acceptable is False:
        fmap_basis, C, hk, zk = mri_exp_approx(
            ti=ti,
            zmap=zmap,
            segments=fieldmap_segments,
            approx_type=approx_type,
            ctest=True,
            **exp_approx_args,
        )
        Eh = xp.exp(-ti[:, None] * zk.T)
        Ep = xp.dot(fmap_basis, C)
        err = xp.abs(Eh - Ep)
        # nhist = len(zk)
        # mse = xp.mean(err**2, axis=0)
        # wrms = xp.sqrt((mse[None, :] * hk[:, None]) / hk.sum())
        # ik = xp.round(xp.linspace(1,nhist,8))
        err_max = xp.max(err.mean())
        print(
            "fieldmap_segments={}, err_max={}".format(
                fieldmap_segments, err_max
            )
        )

        if err_max > err_thresh:
            print("approximation is poor.  increasing fieldmap_segments by 1")
            fieldmap_segments += 1
            fieldmap_segments_adjust += 1  # try again
        else:
            # accuracy is satisfactory as is
            error_acceptable = True

        if fieldmap_segments_adjust > 9:
            raise ValueError(
                "Unable to get reasonable accuracy with small increase in fieldmap_segments"
            )
    return fieldmap_segments


class MRI_Operator(LinearOperatorMulti):
    """MRI_Operator object for non-Cartesian MR image reconstruction.

    Note: This object can also model coil sensitivity maps and relaxation or
    off-resonance effects for MRI.

    create a system matrix object by calling:
    A = MRI_Operator( ... )
    The forward operation is then:
        y = A * x;
    The adjoint operation is then:
        y = A.H * x;

    Intended for use with iterative image reconstruction in MRI.
    """

    __weights = None
    __unitary_scaling = False
    __pixel_basis = "dirac"

    @profile
    def __init__(
        self,
        kspace,
        shape,
        mask=None,
        grid_os_factor=1.5,
        pixel_basis="dirac",
        weights=None,
        spectral_offsets=None,
        on_gpu=False,
        precision="single",
        # dtype=np.complex64,  # not used. precision determines this
        nufft_kwargs={},
        exact=False,
        fov=1,
        n_shift=None,
        exp_approx_args={},
        order="F",
        coil_sensitivities=None,
        unitary_scaling=False,
        squeeze_reps=True,
        verbose=False,
        xp=None,
        **kwargs,
    ):
        """Initialize a non-Cartesian MRI_Operator.

        Parameters
        ----------
        kspace : array_like
            2D array of k-space coordinates :math:`[n_{samples}, n_{dim}]`.
            (in inverse units of fov)
        mask : array_like or None
            boolean mask for the image array
        pixel_basis : {'dirac', 'rect'}, optional
            pixel basis for the continuous-to-discrete representation
        coil_sensitivities : array_like, optional
            2D array of coil sensitivity maps :math:`[n_{voxels}, n_{coils}]`.
        zmap : array_like, optional
            relax_map + 2j*pi*field_map.  real part is relaxation, imaginary
            part is the off-resonance field map.
        ti : array_like, optional
            time vector corresponding to the field map

        Other Parameters
        ----------------
        exp_approx_args : dict, optional
            additional kwargs for `mri_exp_approx`
        nufft_kwargs : dict, optional
            additional kwargs for `NUFFT_Operator`
        exact : bool, optional
            if True use slow, exact transform.  if False (default), use NUFFT
        n_shift : array_like, optional
            image domain shift for exact case
        fov : array_like, optional
            image field of view (e.g. in mm)
        Gnufft : NUFFT_Operator
            use the provided Gnufft operator instead of generating a new one
        fieldmap_segments : int, optional
            number of fieldmap approximation terms (see `mri_exp_approx`)
        acorr_fieldmap_segments : int, optional
            number of fieldmap approximation terms for autocorrelation
            histogram in Toeplitz version
        n_shots : int, optional
            Used to specify that the trajectory is multiple shots over
            identical time intervals.  Used to speed up zmap calculation by
            passing in `ti` for a single repetition.
        dtype : {np.float32, np.float64}
            used to specify the precision

        Attributes
        ----------
        TODO

        Notes
        -----
        any additional kwargs aside from those passed in above get passed onto
        the NUFFT_Operator object

        Update or change the zmap (e.g. for dynamic cases) via:
        A.new_zmap(zmap=zmap, ti=ti)

        Extended from original Matlab implementation:
            by Jeff Fessler, The University of Michigan

        """
        if on_gpu:
            xp = cupy
            kspace = cupy.asarray(kspace)
            loc_in = loc_out = "gpu"
        else:
            xp, on_gpu = get_array_module(kspace, xp=np)
            loc_in = loc_out = "cpu"
        self._on_gpu = on_gpu
        shape = tuple(shape)
        self.ndim = len(shape)
        self.Nd = shape  # TODO: remove?

        if "nufft_args" in kwargs:
            raise ValueError(
                "use of nufft_args is outdated. use nufft_kwargs instead."
            )

        if mask is None:
            self.mask = None
            self.nmask = prod(shape)
        else:
            self.mask = xp.asarray(mask, dtype=bool)
            if self.mask.shape != shape:
                raise ValueError("mask.shape must match the specified shape")
            nmask = xp.count_nonzero(self.mask)
            if self.xp is np:
                self.nmask = nmask
            else:
                self.nmask = nmask.get()

        if verbose:
            print("kwargs keys = {}".format(kwargs.keys()))
            print(kwargs)

        self.precision = precision
        if self.precision == "single":
            self._real_dtype = np.float32
            self._cplx_dtype = np.complex64
        else:
            self._real_dtype = np.float64
            self._cplx_dtype = np.complex128
        self.dtype = self._cplx_dtype
        if self.dtype not in [xp.complex64, xp.complex128]:
            # Note: MUST KEEP OPERATOR DTYPE REAL OR CONJUGATE OF INPUTS AND
            # OUTPUTS WILL BE TAKEN CAUSING ERRONEOUS RESULTS
            # operator type must be defined as (real) float or double
            raise ValueError("unsupported dtype: must be 32 or 64 bit float")

        if kspace.ndim != 2 or kspace.shape[1] > 3:
            raise ValueError(
                "Invalid kspace shape. Must be a 2D array where the second "
                "axis corresponds to 1D-3D spatial axes."
            )
        if kspace.dtype != self._real_dtype:
            kspace = kspace.astype(self._real_dtype, copy=False)
        self.kspace = kspace
        ndim = kspace.shape[1]

        # defaults
        self.exact = exact
        self.fov = fov
        if len(self.fov) == 1 and ndim > 1:
            self.fov = np.asarray(self.fov.tolist() * ndim)

        self.omega = self._kspace_to_omega(kspace)  # set after Nd, fov
        if xp.max(xp.abs(self.omega)) > xp.pi + 1e-6:
            warnings.warn(
                "Warning in MRI_Operator: omega exceeds pi. Was "
                "this intended?"
            )

        if self.exact:
            self.n_shift = n_shift
            if self.n_shift is None:
                self.n_shift = tuple(d // 2 for d in self.shape)
            self.n_shift = tuple(self.n_shift)
        else:
            self.n_shift = None

        # arguments for mri_exp_approx()
        self.exp_approx_args = exp_approx_args

        if order != "F":
            raise ValueError("MRI_Operator only supports order='F'.")
        self.order = order

        # SENSE-stuff
        self.coil_sensitivities = coil_sensitivities
        if self.coil_sensitivities is None:
            self.Ncoils = 1
        else:
            # need Fortran contiguous sensitivity maps
            self.coil_sensitivities = xp.asfortranarray(self.coil_sensitivities)

            if self.coil_sensitivities.ndim != 2:
                raise ValueError(
                    "'coil_sensitivities' must be either be a 2d array "
                    + "[Npixels, Ncoils] or None"
                )
            self.Ncoils = self.coil_sensitivities.shape[1]

        nargin = prod(shape)
        nargout = self.kspace.shape[0] * self.Ncoils
        self.__unitary_scaling = False

        # multispectral reconstruction
        self.spectral_offsets = spectral_offsets
        if self.spectral_offsets is not None:
            self.nspectra = len(self.spectral_offsets)
            self.spectral_offsets = tuple(self.spectral_offsets)
        else:
            self.nspectra = 1
        nargin *= self.nspectra

        #
        # initialize exact transform or NUFFT approximation object
        #
        if self.exact:
            self._init_exact()
        else:

            if on_gpu:
                self.omega = cupy.asarray(self.omega)

            # Find a size amenable to fast FFT, but also a multiple of 8
            # The choice of 8 is somewhat arbitrary, but ensures that for
            # instance a 3-level periodization wavelet transform can be
            # performed in a non-redundant fashion.
            os_grid_shape = tuple(
                [
                    next_fast_len_multiple(ceil(grid_os_factor * s), 8)
                    for s in shape
                ]
            )

            # Default NUFFT parameters
            self.nufft_kwargs = dict(
                Jd=(6,) * self.ndim,
                n_shift=tuple([n / 2 for n in self.Nd]),
                Ld=1024,
                mode="table1",
                phasing="real",
                ortho=False,  # TODO: change default to True?
            )
            if not isinstance(nufft_kwargs, dict):
                raise ValueError("nufft_kwargs must be a dictionary")
            # update defaults using any user-specified NUFFT kwargs
            valid_nufft_keys = list(self.nufft_kwargs)
            keys_to_pop = [
                k for k in nufft_kwargs.keys() if k not in valid_nufft_keys
            ]
            if keys_to_pop:
                warnings.warn(
                    "The following user-provided entries in nufft_kwargs will "
                    f"be ignored: {keys_to_pop}"
                )
                for key in keys_to_pop:
                    nufft_kwargs.pop(key, None)
            self.nufft_kwargs.update(nufft_kwargs)

            if self.n_shift is not None and any(self.n_shift):
                raise ValueError("n_shift ignored in NUFFT-based case")

            if ("squeeze_reps" in self.nufft_kwargs) and self.nufft_kwargs[
                "squeeze_reps"
            ]:
                raise ValueError(
                    "squeeze_reps = True not allowed within nufft_kwargs"
                )

            self.Gnufft = NUFFT_Operator(
                mask=mask,
                omega=self.omega,
                Nd=shape,
                Kd=os_grid_shape,
                precision=self.precision,
                on_gpu=on_gpu,
                loc_in=loc_in,
                loc_out=loc_out,
                squeeze_reps=False,
                **self.nufft_kwargs,
            )
            # store the oversampled image dimensions in the base object as well
            self.Kd = self.Gnufft.Kd

        # configure spatial basis Fourier transform
        self.pixel_basis = pixel_basis

        # configure DCF and/or basis weights
        # (must come after self._init_basis())
        self._init_density_compensation(
            weights, nargout, apply_basis_transform=True
        )

        # configure fieldmap corrected recon
        # This sets self.ti and self.zmap
        self._init_fieldmap(kwargs)

        if self.spectral_offsets is not None:
            # Note: uses self.ti as set by self._init_fieldmap
            self._init_multispectral(
                n_shots=self.kspace.shape[0] // self.ti.size
            )

        """
        select forward and adjoint operators depending on whether transform is
        exact or NUFFT-based and whether computation is to be done on the CPU
        or GPU.
        """
        if self.exact:
            if self.spectral_offsets is not None:
                raise ValueError(
                    "no multispectral implementation for the exact case"
                )
            else:
                matvec = self._forw_exact
                matvec_adj = self._back_exact
        else:
            if self.spectral_offsets is not None:
                matvec = self._forw_multispectral
                matvec_adj = self._back_multispectral
            else:
                matvec = self._forw
                matvec_adj = self._back

        self.__matvec = matvec
        self.__matvec_transp = matvec_adj
        self.__matvec_adj = matvec_adj
        super(MRI_Operator, self).__init__(
            nargin,
            nargout,
            symmetric=False,
            hermetian=False,
            matvec=self.__matvec,
            matvec_adj=self.__matvec_transp,
            matvec_transp=self.__matvec_adj,
            nd_input=False,
            nd_output=False,
            shape_in=(nargin, 1),
            shape_out=(nargout, 1),
            order="F",
            matvec_allows_repetitions=True,
            squeeze_reps=squeeze_reps,
            mask_in=None,  # self.mask,
            mask_out=None,
            dtype=self.dtype,
            loc_in=loc_in,
            loc_out=loc_out,
        )

        # make sure coils sensitivity, fieldmap etc. match the expected dtype
        self._update_array_precision()

        # if requested, scaling the weights to give an operator which is
        # approximately unitary (norm preserving)
        self.unitary_scaling = unitary_scaling

    def _init_basis(self):
        """Called by __init__ to initialize the PixelBasis."""
        self.basis = PixelBasis(
            self.kspace,
            pixel_basis=self.__pixel_basis,
            fov=self.fov,
            Nd=self.Nd,
        )

    @property
    def pixel_basis(self):
        """Return the current PixelBasis object."""
        return self.__pixel_basis

    @pixel_basis.setter
    def pixel_basis(self, pixel_basis):
        """Update the pixel basis to a new pixel_basis."""
        self.__pixel_basis = pixel_basis
        self._init_basis()

    def _init_exact(self):
        """Iniatilization for exact (slow) transform."""
        N = self.Nd
        ndim = self.ndim
        xp = self.xp
        mask = self.mask
        if self.n_shift is None:
            raise ValueError("n_shift required when exact=True")
        if True:
            # TODO: test this n-dimensional refactoring
            nn = []
            for d in range(ndim):
                nn.append(xp.arange(N[d]) - self.n_shift[d])
            nn = xp.meshgrid(*nn, indexing="ij")
            for d in range(ndim):
                self.u[:, d] = self._apply_mask(nn[d], squeeze_output=True)
            self.u = xp.empty((self.nmask, ndim), dtype=self._real_dtype)
        else:
            if len(N) == 2:
                n1, n2 = xp.meshgrid(
                    xp.arange(0, N[0]) - self.n_shift[0],
                    xp.arange(0, N[1]) - self.n_shift[1],
                    indexing="ij",
                )
                self.u = xp.zeros((xp.count_nonzero(mask), 2))
                # transposes to get Fortran style ordering to match Matlab
                self.u[:, 0] = n1.T[mask.T]
                self.u[:, 1] = n2.T[mask.T]
                self.u = self.u.T  # (2, np)
            elif len(N) == 3:
                n1, n2, n3 = xp.meshgrid(
                    xp.arange(0, N[0]) - self.n_shift[0],
                    xp.arange(0, N[1]) - self.n_shift[1],
                    xp.arange(0, N[2]) - self.n_shift[2],
                    indexing="ij",
                )
                self.u = xp.zeros((xp.count_nonzero(mask), 2))
                # transposes to get Fortran style ordering to match Matlab
                self.u[:, 0] = n1.T[mask.T]
                self.u[:, 1] = n2.T[mask.T]
                self.u[:, 2] = n3.T[mask.T]
                self.u = self.u.T  # (3, np)
            else:
                raise ValueError("Only 2D and 3D cases done")
        # self.u = complexify(self.u)
        self.v = 1j * self.omega.T  # (2, np)

    #        # TODO: use basis_transform_op to automatically handle repetitions?
    #             No: for now have folded basis.transform into self.weights
    #        if self.basis.transform is not None:
    #            self.basis_transform_op = DiagonalOperator(self.basis.transform)
    #            if self.Ncoils > 1:
    #                self.basis_transform_op = BlockDiagLinOp(
    #                    [self.basis_transform_op, ] * self.Ncoils)

    # TODO: replace [] with None
    def _init_fieldmap(self, kwargs):
        """Called by __init__ to initialize the fieldmap variables."""
        # new fieldmap stuff
        # n_shots can speed up zmap calculation in case where the
        # trajectory is multiple shots over identical time interval
        xp = self.xp
        self.zmap = kwargs.pop("zmap", None)
        self.ti = kwargs.pop("ti", None)
        if self.ti is not [] and self.ti is not None:
            self.ti = xp.asarray(self.ti)
        self.fieldmap_segments = kwargs.pop("fieldmap_segments", None)
        if (
            self.fieldmap_segments is None
            and self.ti is not None
            and self.zmap is not None
        ):
            self.fieldmap_segments = determine_fieldmap_segments(
                self.ti, self.zmap, fieldmap_segments_start=3, xp=xp
            )
        # acorr_fieldmap_segments is for autocorrelation zmap. default is kwargs['L']
        self.acorr_fieldmap_segments = kwargs.pop(
            "acorr_fieldmap_segments", self.fieldmap_segments
        )
        self.n_shots = kwargs.pop("n_shots", 1)

        # add zmap if available now.
        if self.zmap is not None:
            if (not self.exact) and (
                self.fieldmap_segments is None or self.ti is None
            ):
                raise ValueError(
                    "field-corrected recon:  zmap specified without ti and fieldmap_segments"
                )
            self.new_zmap(
                ti=self.ti,
                zmap=self.zmap,
                fieldmap_segments=self.fieldmap_segments,
                n_shots=self.n_shots,
            )

    def _init_density_compensation(
        self, weights, nargout=None, apply_basis_transform=False
    ):

        # incorporate self.basis.transform into the weights array
        if apply_basis_transform and self.basis.transform is not None:
            # TODO!!!: fix so that this doesn't break _scale_weights!!
            if self.xp.iscomplexobj(self.basis.transform):
                raise ValueError("TODO: complex case not tested")
            if weights is None:
                weights = self.basis.transform.copy()
            else:
                # fold the basis.transform into the weights
                if weights.ndim != 1:
                    raise ValueError("expected a 1D array of weights")
                if weights.size == self.basis.transform.size:
                    weights *= self.basis.transform
                elif weights.size == self.basis.transform.size * self.Ncoils:
                    weights *= self.xp.concatenate(
                        [self.basis.transform] * self.Ncoils
                    )
                else:
                    raise ValueError("invalid size for weights")

        """Initialize Density Compensation weights."""
        if weights is None:
            return None
        if nargout is None:
            nargout = self.nargout
        if weights is not None:
            xp = self.xp
            if xp is np:
                loc_in = loc_out = "cpu"
            else:
                loc_in = loc_out = "gpu"
            if isinstance(weights, xp.ndarray):
                if weights.dtype != self._real_dtype:
                    weights = weights.astype(self._real_dtype)
                weights = DiagonalOperator(
                    weights, loc_in=loc_in, loc_out=loc_out
                )
            elif not isinstance(weights, DiagonalOperator):
                raise ValueError(
                    "weights must be an ndarray or DiagonalOperator"
                )
            if weights.dtype != self._real_dtype:
                # retrieve the diagonal and convert to the desired dtype
                weights = weights.diag.diagonal().astype(self._real_dtype)
                weights = DiagonalOperator(
                    weights, loc_in=loc_in, loc_out=loc_out
                )
            if weights.nargin != nargout:
                if weights.nargin * self.Ncoils == nargout:
                    weights = BlockDiagLinOp(
                        [weights] * self.Ncoils, loc_in=loc_in, loc_out=loc_out
                    )
                else:
                    raise ValueError("mismatch between kspace size and weights")

        self.__weights = weights

    def _init_multispectral(self, n_shots):
        # initialize arrays for multi-spectral reconstruction
        xp = self.xp
        if self.spectral_offsets is None:
            return
        if self.ti is None:
            raise ValueError(
                "Using spectral_offsets requires ti to be specified"
            )
        self.offset_arrays = []
        for f in self.spectral_offsets:
            if f == 0:
                self.offset_arrays.append(None)
            elif np.isscalar(f):
                v = xp.exp(1j * 2 * xp.pi * f * self.ti.ravel(order="F"))
                if n_shots > 1:
                    v = xp.concatenate((v,) * n_shots, axis=0)
                self.offset_arrays.append(v)
            else:  # tuple of 2-tuples
                if not all(len(ff) == 2 for ff in f):
                    raise ValueError(
                        "elements of spectral offsets must be scalar or a "
                        "tuple of 2-tuples where each 2-tuple is a "
                        "(weight, offset) pair"
                    )
                if abs(1.0 - sum(ff[0] for ff in f)) > 1e-2:
                    raise ValueError("sum of the weights should equal 1.0")
                offset_set = []
                for w, ff in f:
                    v = xp.exp(1j * 2 * xp.pi * ff * self.ti.ravel(order="F"))
                    if n_shots > 1:
                        v = xp.concatenate((v,) * n_shots, axis=0)
                    offset_set.append((w, v))
                self.offset_arrays.append(tuple(offset_set))

    def _update_dtype(self, arr, mode=None):
        """ Fixup an object's dtype to match the precision of the oeprator."""
        xp = self.xp
        if mode is None:
            if xp.iscomplexobj(arr):
                if arr.dtype != self._cplx_dtype:
                    arr = arr.astype(self._cplx_dtype)
            else:
                if arr.dtype != self._real_dtype:
                    arr = arr.astype(self._real_dtype)
        elif mode == "real":
            if arr.dtype != self._real_dtype:
                arr = arr.astype(self._real_dtype)
        elif mode == "complex":
            if arr.dtype != self._cplx_dtype:
                arr = arr.astype(self._cplx_dtype)
        else:
            raise ValueError("unrecognized dtype mode")
        return arr

    def _update_array_precision(self):
        # make sure all dtypes are at a consistent precision
        self.kspace = self._update_dtype(self.kspace, "real")
        self.omega = self._update_dtype(self.omega, "real")
        if hasattr(self, "u"):
            self.u = self._update_dtype(self.u)
        if hasattr(self, "v"):
            self.v = self._update_dtype(self.v)
        if self.coil_sensitivities is not None:
            self.coil_sensitivities = self._update_dtype(
                self.coil_sensitivities, "complex"
            )
        if self.zmap is not None:
            self.zmap = self._update_dtype(self.zmap, "complex")
            if hasattr(self, "ti") and self.ti is not None:
                self.ti = self._update_dtype(self.ti, "real")
            if hasattr(self, "fmap_basis") and self.fmap_basis is not None:
                self.fmap_basis = self._update_dtype(self.fmap_basis, "complex")
            if hasattr(self, "fmap_coeffs") and self.fmap_coeffs is not None:
                self.fmap_coeffs = self._update_dtype(
                    self.fmap_coeffs, "complex"
                )
        if self.basis.transform is not None:
            self.basis.transform = self._update_dtype(
                self.basis.transform, "real"
            )

    def _kspace_to_omega(self, kspace):
        xp = self.xp
        ndim = len(self.Nd)
        # configure frequencies in physical units
        omega = xp.empty_like(kspace)
        for id in range(ndim):
            omega[:, id] = (
                2 * xp.pi * kspace[:, id] * self.fov[id] / self.Nd[id]
            )
        if omega.dtype != self._real_dtype:
            omega = omega.astype(self._real_dtype)
        return omega

    @property
    def dim(self):
        """The shape corresponding to the operator."""
        return self.shape

    def _get_weights_array(self):
        xp = self.xp
        if self.weights is None:
            return xp.ones(self.nargout // self.Ncoils, dtype=self._real_dtype)
        elif isinstance(self.weights, BlockDiagLinOp):
            # TODO: manually concatenate blocks? This assumes all are the same
            w = self.weights.blocks[0].diag
        else:
            w = self.weights.diag
        return w

    def _scale_weights(self, power_iterations=15):
        # TODO: can Anorm be a single NUFFT operation instead of all coils?
        xp = self.xp
        cplx_dtype = self._cplx_dtype
        if self.norm is not None:
            lam = power_method(self, dtype=cplx_dtype, niter=power_iterations)
        else:
            Anorm = self.H * self
            lam = power_method(Anorm, dtype=cplx_dtype, niter=power_iterations)
        self._lam = lam
        w = self._get_weights_array()
        new_weights = w / xp.sqrt(lam)
        self._init_density_compensation(
            new_weights,
            # don't apply basis transform again!
            apply_basis_transform=False,
        )
        self.__unitary_scaling = True

    @property
    def weights(self):
        return self.__weights

    @weights.setter
    def weights(self, weights):
        self._init_density_compensation(weights)
        if self.unitary_scaling:
            self._scale_weights()

    @property
    def unitary_scaling(self):
        """Boolean indicating if the norm of the operator is (nearly) unitary.

        i.e. Whether the np.linalg.norm(A.H * A * x) is approximately equal to
        np.linalg.norm(x) for an operator, A. If this is False, and the user
        sets it to True, the appropriately scaling of the ``weights`` property
        is computed via the power method.
        """
        return self.__unitary_scaling

    @unitary_scaling.setter
    def unitary_scaling(self, unitary_scaling):
        """Boolean indicating if the norm of the operator is (nearly) unitary.

        i.e. Whether the np.linalg.norm(A.H * A * x) is approximately equal to
        np.linalg.norm(x) for an operator, A. If the user sets it to True, the
        appropriately scaling of the ``weights`` property is computed via the
        power method.
        """
        if unitary_scaling is True:
            if self.__unitary_scaling is False:
                self._scale_weights()
        elif unitary_scaling is not False:
            raise ValueError("unitary_scaling should be True or False")
        self.__unitary_scaling = unitary_scaling

    def set_unitary_scaling_weights(self, weights):
        """Utility for using user-provided unitary weights instead of computing
        from the power method."""
        self.__weights = weights
        self.__unitary_scaling = True

    def copy_zmap(self, MRIop2):
        """ Copy zmap from an already initialized MRI Operator
        """
        if not hasattr(MRIop2, "zmap") or MRIop2.zmap is None:
            raise ValueError("No zmap to copy")
        # TODO: do some basic dimension checks
        self.ti = MRIop2.ti
        self.zmap = MRIop2.zmap
        self.fmap_coeffs = MRIop2.fmap_coeffs
        self.fieldmap_segments = MRIop2.fieldmap_segments
        self.exp_approx_type = MRIop2.exp_approx_type
        self.exp_approx_args = MRIop2.exp_approx_args
        self.fmap_basis = MRIop2.fmap_basis

    # @profile
    def copy_zmap_new_ti(self, MRIop2, ti, n_shots=1):
        """Used to share zmap, fmap_coeffs across segments in multi-segment recon."""
        if not hasattr(MRIop2, "zmap") or MRIop2.zmap is None:
            raise ValueError("No zmap to copy")

        xp = self.xp
        # make sure precision of dtypes matches the NUFFT operator
        if ti.dtype != self.Gnufft._real_dtype:
            ti = ti.astype(self.Gnufft._real_dtype)

        self.ti = ti.ravel(order="F")
        self.zmap = MRIop2.zmap
        self.fmap_coeffs = MRIop2.fmap_coeffs
        self.fieldmap_segments = MRIop2.fieldmap_segments
        self.exp_approx_type = MRIop2.exp_approx_type
        self.exp_approx_args = MRIop2.exp_approx_args

        # allti_equal = xp.max(self.ti, MRIop2.ti)
        same_ti_vector = False
        if self.ti.shape == MRIop2.ti.shape:
            if xp.max(xp.abs(self.ti - MRIop2.ti)) < 1e-6:
                same_ti_vector = True

        if same_ti_vector:
            self.fmap_basis = MRIop2.fmap_basis
        else:
            self.fmap_basis, C, hk, zk = mri_exp_approx(
                ti=self.ti,
                zmap=self.zmap,
                segments=self.fieldmap_segments,
                approx_type=self.exp_approx_type,
                **self.exp_approx_args,
            )

        # make sure precision of dtypes matches the NUFFT operator
        if self.fmap_basis.dtype != self.Gnufft._cplx_dtype:
            self.fmap_basis = self.fmap_basis.astype(self.Gnufft._cplx_dtype)

        if n_shots > 1:
            """when there are multiple shots with same time vector can just
            duplicate here."""
            self.fmap_basis = xp.tile(self.fmap_basis, [n_shots, 1])

        if xp.any(xp.isnan(self.fmap_basis)):
            raise ValueError("bug: nan values in fmap_basis")

        self.fieldmap_segments = self.fmap_basis.shape[1]

    # @profile
    def new_zmap(
        self,
        ti,
        zmap,
        fieldmap_segments=None,
        n_shots=1,
        acorr_fieldmap_segments=None,
        calc_acorr_fieldmap_basis=False,
        xp=None,
    ):
        """ Initialize the fieldmap approximation basis and coefficients

        Parameters
        ----------
        see main class documentation

        Notes
        -----
        acorr_fieldmap_segments, calc_acorr_fieldmap_basis only required for Toeplitz case
        """
        if xp is None:
            xp = self.xp
        if len(ti) != self.kspace.shape[0] // n_shots:
            raise ValueError("ti size mismatch in new_zmap()")

        # make sure precision of dtypes matches the NUFFT operator
        if not self.exact:
            if ti.dtype != self.Gnufft._real_dtype:
                ti = ti.astype(self.Gnufft._real_dtype)
            if zmap.dtype != self.Gnufft._cplx_dtype:
                zmap = zmap.astype(self.Gnufft._cplx_dtype)
        else:
            pass
            # TODO:  fix dtypes in exact case?

        self.ti = ti.ravel(order="F")

        # apply mask to zmap if it wasn't already masked
        if self.mask is not None:
            # if ((zmap.ndim == self.mask.ndim + 1) and
            #         (zmap.shape[:-1] == self.mask.shape)):
            #     # TODO: trying to allow zmap to have an extra
            #             motion-segmenets dimension
            #     self.zmap = masker(zmap, self.mask).ravel(order=self.order)
            if zmap.shape == self.mask.shape:
                self.zmap = masker(zmap, self.mask, xp=self.xp)
            elif (
                self.xp.count_nonzero(self.mask) == zmap.shape[0]
                and zmap.ndim == 1
            ):
                self.zmap = zmap
            else:
                raise ValueError("zmap size mismatch in new_zmap()")
        else:
            # if zmap.ndim != 1:
            #    raise ValueError("expected a 1D zmap (masked/raveled)")
            self.zmap = zmap.ravel(order=self.order)

        if fieldmap_segments is None:
            # auto-determine an appropriate number of segments
            fieldmap_segments = determine_fieldmap_segments(
                ti, zmap, fieldmap_segments_start=3, xp=xp
            )

        # trick to handle 'exact' case (for which fieldmap_segments is irrelevant)
        if self.exact:
            ndim = len(self.Nd)
            # trick: if already a zmap in u,v, then replace it
            # TODO: check exact case
            # print("self.u.shape = {}".format(self.u.shape))
            # print("self.v.shape = {}".format(self.v.shape))
            # print("self.zmap.shape = {}".format(self.zmap.shape))
            # print("self.ti.shape = {}".format(self.ti.shape))
            self.u = xp.hstack((self.u[:, 0:ndim], self.zmap[:, xp.newaxis]))
            self.v = xp.vstack((self.v[0:ndim, :], self.ti[xp.newaxis, :]))
            return

        if not fieldmap_segments:  # 4th argument is optional, so defaults here:
            if self.fieldmap_segments:
                fieldmap_segments = self.fieldmap_segments
            else:
                raise ValueError(
                    "user must provide self.fieldmap_segments or fieldmap_segments input"
                )

        # initialize exponential approximations for field-corrected
        # reconstruction
        if xp.any(self.zmap.real):
            self.exp_approx_type = ("hist,time,unif", [40, 10])
        else:
            # zmap_range = max(imag(self.zmap(:)))-min(imag(self.zmap(:)));
            # grl_Nbins = max(round(zmap_range/20),40);  %force enough bins to
            # keep at least 10 Hz resolution.  always use at least 40 bins
            grl_Nbins = [40]  # TODO: remove hardcode
            self.exp_approx_type = ("hist,time,unif", grl_Nbins)

        self.fmap_basis, C, hk, zk = mri_exp_approx(
            ti=self.ti,
            zmap=self.zmap,
            segments=fieldmap_segments,
            approx_type=self.exp_approx_type,
            **self.exp_approx_args,
        )

        # make sure precision of dtypes matches the NUFFT operator
        if self.fmap_basis.dtype != self.Gnufft._cplx_dtype:
            self.fmap_basis = self.fmap_basis.astype(self.Gnufft._cplx_dtype)

        if C.dtype != self.Gnufft._cplx_dtype:
            C = C.astype(self.Gnufft._cplx_dtype)

        if n_shots > 1:
            """when there are multiple shots with same time vector can just
            duplicate here."""
            # Note: on CPU this is explicitly tiled
            #       on GPU to save memory only 1 repetition will be stored
            self.fmap_basis = xp.tile(self.fmap_basis, [n_shots, 1])

        if xp.any(xp.isnan(self.fmap_basis)):
            raise ValueError("bug: nan values in fmap_basis")

        self.fmap_coeffs = C.T

        if (
            not acorr_fieldmap_segments
        ):  # 5th argument is optional, so defaults here:
            if self.acorr_fieldmap_segments:
                acorr_fieldmap_segments = self.acorr_fieldmap_segments
            # trick: use "found" L because acorr_fieldmap_segments >= fieldmap_segments generally
            elif isinstance(fieldmap_segments, (list, xp.ndarray, tuple)):
                acorr_fieldmap_segments = [
                    self.fmap_basis.shape[1],
                    fieldmap_segments[1],
                ]  # [fieldmap_segments, rmsmax]
            else:
                acorr_fieldmap_segments = self.fmap_basis.shape[1]

        # store size (seems redundant, but OK)
        self.fieldmap_segments = self.fmap_basis.shape[1]
        if isinstance(fieldmap_segments, (list, xp.ndarray, tuple)):
            self.rmsmax = fieldmap_segments[1]
            print("fieldmap_segments=%d found" % self.fieldmap_segments)

        if calc_acorr_fieldmap_basis:
            # TODO: untested
            # generate one with auto-correlation histogram too.
            # only if no relaxation map!
            if not xp.any(self.zmap.real):
                self.acorr_fieldmap_basis, C, hk, zk = mri_exp_approx(
                    ti=self.ti,
                    zmap=self.zmap,
                    segments=fieldmap_segments,
                    approx_type=self.exp_approx_type,
                    acorr=True,
                    **self.exp_approx_args,
                )

                # trick: for symmetric histogram, it should be real!
                if xp.any(self.acorr_fieldmap_basis.imag > 1e-6):
                    warnings.warn(
                        "imaginary component to self.acorr_fieldmap_basis in "
                        "MRI_Operator"
                    )
                self.acorr_fieldmap_basis = self.acorr_fieldmap_basis.real

                # make sure precision of dtypes matches the NUFFT operator
                if self.acorr_fieldmap_basis.dtype != self.Gnufft._real_dtype:
                    self.acorr_fieldmap_basis = self.acorr_fieldmap_basis.astype(
                        self.Gnufft._real_dtype
                    )

                if C.dtype != self.Gnufft._cplx_dtype:
                    C = C.astype(self.Gnufft._cplx_dtype)

                self.acorr_fieldmap_coeffs = C.T

                self.acorr_fieldmap_segments = self.acorr_fieldmap_basis.shape[
                    1
                ]
                if isinstance(
                    acorr_fieldmap_segments, (list, xp.ndarray, tuple)
                ):
                    print(
                        "acorr_fieldmap_segments=%d found"
                        % self.acorr_fieldmap_segments
                    )
            else:
                raise ValueError("zmap with real component unsupported")

    # @profile
    def _forw_exact(self, x):
        """ exact forward operation on the CPU """
        # TODO: need copy here to avoid modifying original data?
        xp = self.xp
        x = xp.asarray(x).copy()
        x = x.reshape((self.nargin, -1), order=self.order)

        if x.shape[0] != self.shape[1]:
            # [(N),(nc)] to [*N,nc]  #TODO: test
            x = xp.reshape(x, (xp.prod(self.Nd), -1), order=self.order)
            if self.mask is not None:
                x = x[self.mask.ravel(order=self.order), :]  # [np,*nc]

        x = complexify(x)  # force at least 64-bit complex

        y = exp_xform(x, self.u, self.v, xp=self.xp)  # complexify(single(...))
        y = xp.asarray(y)
        if y.ndim == 1:
            y = y[:, xp.newaxis]

        if self.weights:
            # apply density compensation
            # (self.weights incorporates self.basis.transform if necessary)
            y = self.weights * y

        y = xp.asfortranarray(y)  # TODO: necessary?
        return y  # returns an 2D Nd array

    # @profile
    def _apply_mask(self, x, order="F", squeeze_output=False):
        kwargs = dict(order=order, squeeze_output=squeeze_output, xp=self.xp)
        if self.mask is None:
            return x.reshape((self.nmask, -1), order=order)
        else:
            return masker(x, self.mask, **kwargs)

    @profile
    def _forw(self, x, legacy_loop=False):
        """ NUFFT-based forward operation on the CPU """
        xp = self.xp
        # TODO: need copy to avoid modifying original data?
        # x = xp.asarray(x).copy()  # TODO: check and only do this when necessary
        x = x.reshape((self.nargin // self.nspectra, -1), order=self.order)

        nreps = x.shape[-1]  # repetitions
        nt = self.shape[0] // self.Ncoils
        x = complexify(x)  # force at least 64-bit complex

        sn_outside_loop = True
        if sn_outside_loop:
            sn_tmp = self.Gnufft.sn.copy()
            # apply scale factors outside of fieldmap loop
            x = self._apply_mask(sn_tmp, squeeze_output=False) * x
            # TODO: make sure this is valid for the Gnufft object
            self.Gnufft.sn = None
        try:
            if self.zmap is None:  # ) or (not xp.any(self.zmap)):
                loop_over_coils = False
                if loop_over_coils:
                    if (
                        self.coil_sensitivities is not None
                    ):  # and xp.any(self.coil_sensitivities):
                        y = xp.empty(
                            (self.shape[0], nreps),
                            dtype=x.dtype,
                            order=self.order,
                        )
                        for cc in range(self.Ncoils):
                            # TODO: using broadcasting across nreps now instead
                            # of explicit tile of the sensitivities
                            y[cc * nt : (cc + 1) * nt, :] = self.Gnufft * (
                                self.coil_sensitivities[:, cc][:, xp.newaxis]
                                * x
                            )
                    else:
                        y = self.Gnufft * x
                else:
                    # x = xp.empty((self.shape[1]//self.nspectra, nreps),
                    #              dtype=y.dtype, order=self.order)
                    # multi-thread only across coils and repetitions.  this way is
                    # thread safe
                    if self.Ncoils > 1:
                        # put all coils (and/or repetitions) along the 2nd
                        # dimension
                        y_shape = (nt, self.Ncoils * nreps)
                    else:
                        y_shape = (self.shape[0], nreps)
                    y = xp.empty(y_shape, dtype=x.dtype, order=self.order)

                    for repIDX in range(nreps):
                        rep_slice = slice(
                            repIDX * self.Ncoils, (repIDX + 1) * self.Ncoils
                        )
                        if self.coil_sensitivities is not None:
                            tmp = (
                                self.coil_sensitivities
                                * x[:, repIDX : repIDX + 1]
                            )
                            y[:, rep_slice] = self.Gnufft * tmp
                        else:
                            xsl = x[:, repIDX]
                            tmp = self.Gnufft * xsl
                            y[:, rep_slice] = tmp
                    y = y.reshape((self.shape[0], nreps), order=self.order)

            else:  # approximation

                if self.coil_sensitivities is None:
                    # fieldmap, but no coil sensitivities
                    for ll in range(self.fieldmap_segments):
                        # fmap_coeffs will be broadcast when x.shape[1]>1
                        # fmap_basis will be broadcast when x.shape[1]>1
                        tmp = self.fmap_coeffs[:, ll][:, xp.newaxis] * x
                        tmp = self.Gnufft * tmp
                        tmp = self.fmap_basis[:, ll][:, xp.newaxis] * tmp
                        if ll == 0:
                            y = tmp
                        else:
                            y += tmp
                # not None and xp.any(self.coil_sensitivities):
                else:
                    loop_over_coils = True
                    if loop_over_coils:
                        # multiple coils and fieldmap
                        y = xp.empty(
                            (self.shape[0], nreps),
                            dtype=x.dtype,
                            order=self.order,
                        )
                        for ll in range(self.fieldmap_segments):
                            cx = self.fmap_coeffs[:, ll : ll + 1] * x
                            for cc in range(self.Ncoils):
                                # fmap_coeffs will be broadcast when x.shape[1]>1
                                tmp = (
                                    self.coil_sensitivities[:, cc : cc + 1] * cx
                                )
                                tmp = self.Gnufft * tmp
                                # fmap_basis will be broadcast when x.shape[1]>1
                                tmp = self.fmap_basis[:, ll : ll + 1] * tmp
                                if ll == 0:
                                    y[cc * nt : (cc + 1) * nt, :] = tmp
                                else:
                                    y[cc * nt : (cc + 1) * nt, :] += tmp
                    else:
                        # loop is over repetitions instead of coils
                        if self.Ncoils > 1:
                            # put all coils (and/or repetitions) along the 2nd
                            # dimension
                            y_shape = (nt, self.Ncoils * nreps)
                        else:
                            y_shape = (self.shape[0], nreps)
                        y = xp.empty(y_shape, dtype=x.dtype, order=self.order)
                        for ll in range(self.fieldmap_segments):
                            cx = self.fmap_coeffs[:, ll : ll + 1] * x
                            for repIDX in range(nreps):
                                rep_slice = slice(
                                    repIDX * self.Ncoils,
                                    (repIDX + 1) * self.Ncoils,
                                )
                                tmp = (
                                    self.coil_sensitivities
                                    * cx[:, repIDX : repIDX + 1]
                                )
                                tmp = self.Gnufft * tmp
                                y[:, rep_slice] = (
                                    self.fmap_basis[:, ll : ll + 1] * tmp
                                )
                        y = y.reshape((self.shape[0], nreps), order=self.order)
        finally:
            # place the scale factors back into the NUFFT object
            if sn_outside_loop:
                self.Gnufft.sn = sn_tmp

        y = xp.asarray(y)

        if self.weights is not None:
            # apply density compensation
            # (self.weights incorporates self.basis.transform if necessary)
            y = self.weights * y

        if y.ndim == 1:
            y = y[:, xp.newaxis]
        y = xp.asfortranarray(y)  # TODO: necessary?

        return y  # returns an 2D Nd array

    # @profile
    def _forw_multispectral(self, x):
        xp = self.xp
        x = xp.asarray(x).copy()
        x = x.reshape(
            (self.nargin // self.nspectra, self.nspectra, -1), order=self.order
        )

        for n, (f, v) in enumerate(  # TODO: can remove f here
            zip(self.spectral_offsets, self.offset_arrays)
        ):
            if v is not None:
                use_linops = False
                if use_linops:
                    Omega_shift = DiagonalOperator(
                        v,
                        order=self.order,
                        squeeze_reps=False,
                        loc_in=self.loc_in,
                        loc_out=self.loc_out,
                    )
                    if self.Ncoils > 1:
                        Omega_shift = BlockDiagLinOp(
                            (Omega_shift,) * self.Ncoils,
                            squeeze_reps_out=False,
                            loc_in=self.loc_in,
                            loc_out=self.loc_out,
                        )
                    tmp = self._forw(x[:, n, :])
                    if n == 0:
                        y = Omega_shift * tmp
                    else:
                        y += Omega_shift * tmp
                else:
                    # previously this case was necessary because BlockDiagLinOp
                    # had poor performance. That has since been fixed, so
                    # can just switch back to the use_linops=True case.
                    tmp = self._forw(x[:, n, :])
                    tmp_shape = tmp.shape
                    tmp = tmp.reshape(
                        (tmp.shape[0] // self.Ncoils, -1), order=self.order
                    )
                    if isinstance(v, tuple):
                        #                        if n > 0:
                        #                            y = y.reshape(tmp.shape, order=self.order)

                        # weighted sum of frequencies
                        #     (e.g. multispectral fat model)
                        tmp0 = tmp.copy()
                        for noff, (weight, vf) in enumerate(v):
                            if vf.ndim == 1:
                                vf = vf[:, np.newaxis]
                            tmp = (weight * vf) * tmp0
                            tmp = tmp.reshape(tmp_shape, order=self.order)
                            if n == 0 and noff == 0:
                                y = tmp
                            else:
                                y += tmp
                    else:
                        # single frequency with weight 1.0
                        if v.ndim == 1:
                            v = v[:, np.newaxis]
                        tmp = v * tmp
                        tmp = tmp.reshape(tmp_shape, order=self.order)
                        if n == 0:
                            y = tmp
                        else:
                            y += tmp
            else:
                if n == 0:
                    y = self._forw(x[:, n, :])
                else:
                    y += self._forw(x[:, n, :])

        if y.ndim == 1:  # TODO: necessary?
            y = y[:, xp.newaxis]
        y = xp.asfortranarray(y)  # TODO: necessary?
        return y

    # @profile
    def _back_exact(self, y):
        """ exact adjoint operation on the CPU """
        # TODO: need copy here to avoid modifying original data?
        xp = self.xp
        y = xp.asarray(y).copy()
        y = y.reshape((self.nargout, -1), order=self.order)
        if self.weights is not None:
            # apply density compensation
            y = self.weights * y

        y = complexify(y)  # force at least 64-bit complex

        # trick: conj(exp(-uv)) = exp(-conj(u) conj(v))
        vc = xp.conj(self.v)
        uc = xp.conj(self.u)
        x = exp_xform(y, vc, uc, xp=self.xp)
        # x = double6(x);
        x = xp.asarray(x)
        if x.ndim == 1:
            x = x[:, xp.newaxis]
        x = xp.asfortranarray(x)  # TODO: necessary?
        return x  # returns an 2D Nd array

    #
    #    if False:
    #        nreps = 1
    #        y = DiagonalOperator(w) * dtmp.reshape((-1, nreps), order='F')

    @profile
    def _back(self, y):
        """ NUFFT-based adjoint operation on the CPU """
        #                # coils should also be stacked within first dimension
        #        # any additional dimensions should be stacked into the second
        #        if (y.ndim > 2) or (y.shape[0] != self.nargin):
        #            x = x.reshape((nargin, -1), order='F')
        # need copy here to avoid modifying original data?
        xp = self.xp
        y = xp.asarray(y).copy()
        y = y.reshape((self.nargout, -1), order=self.order)
        nreps = y.shape[-1]
        nt = self.shape[0] // self.Ncoils  # self.kspace.shape[0]
        if self.weights is not None:
            # apply density compensation
            y = self.weights * y

        y = complexify(y)  # force at least 64-bit complex

        sn_outside_loop = True
        if sn_outside_loop:
            sn_tmp = self.Gnufft.sn.copy()
            self.Gnufft.sn = None
        try:
            if self.zmap is None:  # or (not xp.any(self.zmap)):
                x = xp.empty(
                    (self.shape[1] // self.nspectra, nreps),
                    dtype=y.dtype,
                    order=self.order,
                )
                # multi-thread only across coils and repetitions.  this way is
                # thread safe
                if self.Ncoils > 1:
                    # put all coils (and/or repetitions) along the 2nd
                    # dimension
                    y = y.reshape((nt, -1), order=self.order)

                y = self.Gnufft.T * y

                if y.ndim == 1:
                    y = y[..., xp.newaxis]

                if nreps == 1:
                    if self.coil_sensitivities is not None:
                        y *= xp.conj(self.coil_sensitivities)
                        x = y.sum(axis=1, keepdims=True)
                    else:
                        x = y
                else:
                    for repIDX in range(nreps):
                        rep_slice = slice(
                            repIDX * self.Ncoils, (repIDX + 1) * self.Ncoils
                        )
                        if self.coil_sensitivities is not None:
                            y[:, rep_slice] *= xp.conj(self.coil_sensitivities)
                            x[:, repIDX] = xp.sum(y[:, rep_slice], axis=1)
                        else:
                            x[:, repIDX] = y[:, rep_slice][:, 0]
            else:  # with fieldmap approximation
                if self.coil_sensitivities is None:  # ) or
                    # (not xp.any(self.coil_sensitivities))):
                    # fieldmap, but no coil sensitivities
                    y = y.reshape((nt, nreps), order=self.order)
                    for ll in range(0, self.fieldmap_segments):
                        # tmp = self.fmap_basis[:, ll][:, xp.newaxis] * y
                        tmp = xp.conj(self.fmap_basis[:, ll : ll + 1]) * y
                        tmp = self.Gnufft.T * tmp
                        if tmp.ndim == 1:
                            tmp = tmp[:, np.newaxis]
                        tmp *= xp.conj(self.fmap_coeffs[:, ll : ll + 1])
                        if ll == 0:
                            x = tmp
                        else:
                            x += tmp
                    del tmp
                else:
                    # fieldmap and coil sensitivities case
                    x = xp.empty(
                        (self.shape[1] // self.nspectra, nreps),
                        dtype=y.dtype,
                        order=self.order,
                    )
                    # multi-thread only across coils and repetitions.  this way is
                    # thread safe
                    if self.Ncoils > 1:
                        # put all coils (and/or repetitions) along the 2nd dim
                        y = y.reshape((nt, -1), order=self.order)
                    else:
                        # even with 1 coil need 2d for proper broadcasting
                        y = y.reshape((nt, 1), order=self.order)

                    xtmp = 0
                    for ll in range(self.fieldmap_segments):
                        tmp = xp.conj(self.fmap_basis[:, ll : ll + 1]) * y
                        tmp = self.Gnufft.T * tmp
                        if tmp.ndim == 1:
                            tmp = tmp[:, np.newaxis]
                        # can broadcast across both coils & repetitions
                        tmp *= xp.conj(self.fmap_coeffs[:, ll : ll + 1])
                        xtmp += tmp
                    del tmp
                    for repIDX in range(nreps):
                        rep_slice = slice(
                            repIDX * self.Ncoils, (repIDX + 1) * self.Ncoils
                        )
                        xtmp[:, rep_slice] *= xp.conj(self.coil_sensitivities)
                        x[:, repIDX] = xp.sum(xtmp[:, rep_slice], axis=1)
                    del xtmp
            if sn_outside_loop:
                if x.ndim > 2:
                    raise ValueError("unexpected ndim for x")
                # apply rolloff correction
                x *= xp.conj(
                    self._apply_mask(sn_tmp, squeeze_output=x.ndim == 1)
                )
        finally:
            if sn_outside_loop:
                # store rolloff factors back in Gnufft object
                self.Gnufft.sn = sn_tmp

        x = xp.asarray(x)
        if x.ndim == 1:
            x = x[:, xp.newaxis]
        # x = xp.asfortranarray(x)  # TODO: necessary?
        return x  # returns an 2D Nd array

    # @profile
    def _back_multispectral(self, y):
        xp = self.xp
        nreps = y.size // self.nargout
        x = xp.empty(
            (self.shape[1] // self.nspectra, self.nspectra, nreps),
            dtype=y.dtype,
            order=self.order,
        )
        for n, (f, v) in enumerate(
            zip(self.spectral_offsets, self.offset_arrays)
        ):
            if v is not None:
                use_linops = False
                if use_linops:
                    Omega_shift = DiagonalOperator(
                        v,
                        order=self.order,
                        squeeze_reps=False,
                        loc_in=self.loc_in,
                        loc_out=self.loc_out,
                    )
                    if self.Ncoils > 1:
                        Omega_shift = BlockDiagLinOp(
                            (Omega_shift,) * self.Ncoils,
                            squeeze_reps_out=False,
                            loc_in=self.loc_in,
                            loc_out=self.loc_out,
                        )
                    tmp = Omega_shift.H * y
                else:
                    y_shape = y.shape
                    y = y.reshape(
                        (self.shape[0] // self.Ncoils, -1), order=self.order
                    )
                    if isinstance(v, tuple):
                        # weighted sum of frequencies
                        #     (e.g. multispectral fat model)
                        for noff, (weight, vf) in enumerate(v):
                            # single frequency with weight = 1.0
                            if vf.ndim == 1:
                                vf = vf[:, np.newaxis]
                            if noff == 0:
                                tmp = xp.conj(weight * vf) * y
                            else:
                                tmp += xp.conj(weight * vf) * y
                    else:
                        # single frequency with weight = 1.0
                        if v.ndim == 1:
                            v = v[:, np.newaxis]
                        tmp = xp.conj(v) * y
                    tmp = tmp.reshape(y_shape, order=self.order)
                x[:, n, :] = self._back(tmp)
            else:
                x[:, n, :] = self._back(y)
        if x.ndim == 1:  # TODO: necessary?
            x = x[:, xp.newaxis]
        x = xp.asfortranarray(x)  # TODO: necessary?
        return x

    def prep_Toeplitz(self, weights=None):
        raise NotImplementedError("TODO")
        xp = self.xp
        if weights is not None:
            if self.weights is not None:
                raise ValueError("TODO?")
            if not isinstance(weights, DiagonalOperator):
                W_op = DiagonalOperator(weights, order="F")
            else:
                W_op = weights
        else:
            if self.weights is not None:
                raise ValueError("TODO")  # copy the existing weights
            else:
                W_op = None
            # raise ValueError("weights must be a DiagonalOperator")

        # TODO: update for basis.transform built into self.weights
        if self.basis.transform is not None:
            basis_sq = xp.real(
                xp.conj(self.basis.transform) * self.basis.transform
            )
            if W_op is not None:
                # fold the basis transform into the DCF weights
                W_op.diag.setdiag(W_op.diag.diagonal() * basis_sq)
            else:
                W_op = DiagonalOperator(basis_sq, order="F")
        self.W_op = W_op

        if not hasattr(self.Gnufft, "Q"):
            self.Gnufft.prep_toeplitz()

        # if False:
        #     if self.W_op is None:
        #         T = CompositeLinOp(self.Gnufft.H, self.Gnufft)
        #     else:
        #         T = CompositeLinOp(self.Gnufft.H, W_op, self.Gnufft)
        # else:
        #     # FT_op = FFT_Operator(self.Nd)
        #     pass

    @profile
    def norm(self, x):
        """implementation of the normal operator (A.T*A)."""
        x = complexify(x, xp=self.xp)
        # if self.mask is not None:
        #     x = xp.embed(x, self.mask, xp=self.xp)
        # try:
        #     if x.size % self.nargin != 0:
        #         raise ValueError("wrong size input")
        #     nreps = x.size // self.nargin
        # except IndexError:
        #     nreps = 1

        if hasattr(self, "Q"):
            #   D = self.weights
            raise NotImplementedError("TODO")
        #            slices = [slice(None), ]*x.ndim
        #            for d in range(len(self.Nd)):
        #                slices[d] = slice(self.Nd[d])
        #            if nreps == 1:
        #                y = fftn(x, s=self.Kd)
        #                y *= self.Q
        #                y = ifftn(y)[slices]
        #            else:
        #                x = x.reshape(tuple(self.Nd) + (nreps, ),
        #                              order=self.order)
        #                fft_axes = tuple(xp.arange(len(self.Nd)))
        #                y = fftn(x, s=self.Kd, axes=fft_axes)
        #                y *= self.Q[..., xp.newaxis]  # add an axis for repetitions
        #                y = ifftn(y, axes=fft_axes)[slices]
        else:
            y = self * x
            y = self.H * y
        return y



"""Functions related to Non-Cartesian MRI sampling density compensation.

See also the separate ``mrrt.sdc3`` module for a more well-tested algorithm.
"""
import warnings

import numpy as np


from mrrt.nufft._nufft import nufft_forward, nufft_adj

# TODO: add docstring and literature reference


def mri_dcf_pipe(nufft_op, niter=8, thresh=0.02, verbose=False):
    from mrrt.mri.operators import MRI_Operator, NUFFT_Operator

    if isinstance(nufft_op, MRI_Operator):
        nufft_op = nufft_op.Gnufft
    elif not isinstance(nufft_op, NUFFT_Operator):
        raise ValueError("Gnufft_op must be an MRI_Operator or NUFFT_Operator")
    xp = nufft_op.xp
    wi = xp.ones(nufft_op.shape[0], dtype=nufft_op._cplx_dtype)
    itr = 0
    goal = np.inf
    max_diff = xp.max(xp.abs(goal - 1))
    while max_diff > thresh:
        itr += 1
        goal = nufft_adj(nufft_op, xk=wi, grid_only=True)
        goal = nufft_forward(nufft_op, x=goal, grid_only=True)
        max_diff = xp.max(xp.abs(goal - 1))
        if verbose:
            print("iteration {}, max_err = {}".format(itr, max_diff))
        goal.shape = wi.shape
        wi = wi / goal.real
        if itr > niter:
            warnings.warn(
                "thresh not reached prior to {} iterations".format(niter)
            )
            break
    return wi.real



from numpy.testing import Tester

from ._density_compensation import *  # noqa
from ._partial_fourier import *  # noqa
from ._unwrap import *  # noqa
from .coils import *  # noqa
from .field_maps import *  # noqa
from .sim import *  # noqa
from .version import __version__  # noqa

test = Tester().test



"""Coil compression, pre-whitening and sensitivy map estimation utilities.

Much of this code is modified from idmrmrd-python-tools
https://github.com/ismrmrd/ismrmrd-python-tools

Much code that was 2D only has been updated to work with 3D (or nD) data.
The user can specify which axis of the array corresponds to the coils.

The function calculate_csm_inati_iter is an adaptation of a coil sensitivity
map estimation function by Souheil Inati that is included in Gadgetron.

"""

import warnings
import numpy as np
from mrrt.utils import reale, get_array_module


__all__ = ["apply_pca_weights", "coil_pca", "coil_pca_noncart"]


def _pca_inner(caldata, neig, percentile=None, verbose=False, xp=None):
    # caldata must be shape (ncal, ncoils)
    xp, on_gpu = get_array_module(caldata, xp)
    if neig is None:
        neig = caldata.shape[-1]
    elif neig > caldata.shape[-1]:
        raise ValueError(
            "number of eigencoils cannot exceed the number of original coils"
        )

    cov_mat = xp.dot(xp.conj(caldata.T), caldata)
    # cov_mat is Hermitian symmetric so use the faster eigh instead of eig
    if xp is np:
        w, v = np.linalg.eigh(cov_mat)
    else:
        # can only run eigh on the CPU at the moment
        w, v = np.linalg.eigh(cov_mat.get())
        w = xp.asarray(w)
        v = xp.asarray(v)

    w = reale(w)
    si = xp.argsort(w)[::-1]  # largest to smallest
    if verbose:
        from matplotlib import pyplot as plt

        plt.figure()
        plt.subplot(121)
        if on_gpu:
            plt.plot((w[si] / xp.sum(w[si])).get(), "k.-")
        else:
            plt.plot(w[si] / xp.sum(w[si]), "k.-")
        plt.ylabel("normalized eigenvalue")
        plt.xlabel("channel (sorted by descending eigenvalue)")
        plt.subplot(122)
        x = np.arange(1, caldata.shape[-1] + 1)
        if on_gpu:
            plt.plot(x, xp.cumsum(w[si] / xp.sum(w[si])).get(), "k.-")
        else:
            plt.plot(x, xp.cumsum(w[si] / xp.sum(w[si])), "k.-")
        plt.plot([x[0], x[-1]], [0.95, 0.95], "k--")
        plt.plot([x[0], x[-1]], [0.98, 0.98], "k:")
        plt.plot([neig, neig], [0, 1], "r:")
        plt.ylabel("normalized eigenvalue cumsum")
        plt.ylabel("# of channels")
        plt.axis("tight")

    v = v[:, si]  # eigenvectors sorted
    if percentile is None:
        return v
    else:
        if percentile < 90:
            warnings.warn("percentile < 90%: was this intentional?")
        neig = xp.where(xp.cumsum(w[si] / xp.sum(w[si])) >= percentile / 100)[
            0
        ][0]
        return v, neig


def apply_pca_weights(data, pca_matrix, neig, coil_axis=-1, xp=None):
    xp, on_gpu = get_array_module(data, xp)
    if coil_axis != -1:
        # coil axis must come last
        data = xp.swapaxes(data, -1, coil_axis)
    s = data.shape[:-1] + (neig,)
    data = xp.dot(data, pca_matrix[:, :neig])
    data = data.reshape(s, order="F")
    if coil_axis != -1:
        # swap coil axis back to original location
        data = xp.swapaxes(data, -1, coil_axis)
    return data


def coil_pca_noncart(
    data,
    kabs=None,
    rkeep=1,
    coil_axis=-1,
    neig=12,
    verbose=False,
    pca_matrix_only=False,
    xp=None,
):
    xp, on_gpu = get_array_module(data, xp)
    if coil_axis != -1:
        # coil axis must come last
        data = xp.swapaxes(data, -1, coil_axis)

    ncoils = data.shape[-1]
    if ncoils < neig:
        raise ValueError("too few coils")

    if rkeep < 1:
        if data.ndim != 2:
            raise ValueError("expected a 2D array of data [nsamples, ncoils]")

        if kabs is None:
            raise ValueError("rkeep option requires kspace magnitude vector")

        cal_keep = xp.where(kabs <= rkeep)
        caldata = data[cal_keep[0], :]
        s = None
    else:
        s = data.shape
        caldata = data.reshape((-1, ncoils), order="F")
    pca_mtx = _pca_inner(caldata, neig=neig, verbose=verbose)
    if pca_matrix_only:
        return pca_mtx

    data = apply_pca_weights(data, pca_matrix=pca_mtx, neig=neig)
    if s is not None:
        s = list(s)
        s[-1] = neig
        data = data.reshape(s, order="F")

    if coil_axis != -1:
        # swap coil axis back to original location
        data = xp.swapaxes(data, -1, coil_axis)
    return data


def coil_pca(
    data,
    ncal_x=64,
    ncal_y=64,
    ncal_z=64,
    coil_axis=-1,
    neig=None,
    percentile=100,
    verbose=True,
    pca_matrix_only=False,
    xp=None,
):
    xp, on_gpu = get_array_module(data, xp)
    ncoils = data.shape[coil_axis]
    #    if neig >= ncoils:
    #        return data

    if coil_axis != -1:
        # coil axis must come last
        data = xp.swapaxes(data, -1, coil_axis)

    ncal_y = min(data.shape[0], ncal_x)
    ncal_x = min(data.shape[1], ncal_y)
    ncal_z = min(data.shape[2], ncal_z)

    if data.ndim == 4:
        nx, ny, nz = data.shape[:-1]
    else:
        nx, ny = data.shape[:-1]
    nx_2 = nx // 2
    ny_2 = ny // 2
    if data.ndim == 4:
        nz_2 = nz // 2

    if ncal_x == -1:
        slice_x = slice(None)
    else:
        slice_x = slice(nx_2 - ncal_x // 2, nx_2 + ncal_x // 2)
    if ncal_y == -1:
        slice_y = slice(None)
    else:
        slice_y = slice(ny_2 - ncal_y // 2, ny_2 + ncal_y // 2)

    if data.ndim == 4:
        if ncal_z == -1:
            slice_z = slice(None)
        else:
            slice_z = slice(nz_2 - ncal_z // 2, nz_2 + ncal_z // 2)
        cal_slices = [slice_x, slice_y, slice_z, slice(None)]
    else:
        cal_slices = [slice_x, slice_y, slice(None)]

    caldata = data[cal_slices].reshape((-1, ncoils), order="F")
    data = data.reshape((-1, ncoils), order="F")
    if False:
        caldata_demean = caldata - caldata.mean(-1)[:, xp.newaxis]
        U, s, pca_mtx = xp.linalg.svd(caldata_demean, full_matrices=False)
    else:
        pca_mtx, neig = _pca_inner(
            caldata, neig, percentile=percentile, verbose=verbose
        )
        if pca_matrix_only:
            return pca_mtx, neig
        data = apply_pca_weights(data, pca_mtx, neig)
    if coil_axis != -1:
        # swap coil axis back to original location
        data = xp.swapaxes(data, -1, coil_axis)
    return data



import time
import numpy as np

# from pyir.utils import max_percent_diff
from mrrt.mri import mri_exp_mult
from mrrt.utils import config

if config.have_cupy:
    import cupy


def bench_mri_exp_mult(verbose=True):
    """
    test_mri_exp_mult
    """
    L = 1000
    N = 3000
    M = 200
    rstate = np.random.RandomState(0)
    for dtype in [np.float32, np.float64]:
        if dtype == np.float32:
            atol = rtol = 1e-3
        else:
            atol = rtol = 1e-12
        A = rstate.randn(N, L).astype(dtype)
        A = A + 1j * rstate.randn(N, L).astype(dtype)
        ur = rstate.randn(N).astype(dtype)
        ui = rstate.randn(N).astype(dtype)
        vr = rstate.randn(M).astype(dtype)
        vi = rstate.randn(M).astype(dtype)
        u = ur.astype(dtype) + 1j * ui.astype(dtype)
        v = vr.astype(dtype) + 1j * vi.astype(dtype)

        # Test with ur, v
        tstart = time.time()
        d1 = mri_exp_mult(A, ur, v)
        t_python = time.time() - tstart

        if verbose:
            print("duration (python) = {} s".format(t_python))

        if config.have_cupy:

            tstart = time.time()
            Ag = cupy.asarray(A)
            urg = cupy.asarray(ur)
            vg = cupy.asarray(v)
            d2 = mri_exp_mult(Ag, urg, vg).get()
            t_gpu = time.time() - tstart
            cupy.testing.assert_allclose(d1, d2, atol=atol, rtol=rtol)

            tstart = time.time()
            d3 = mri_exp_mult(Ag, urg, vg)
            t_gpu_single2 = time.time() - tstart
            cupy.testing.assert_allclose(d1, d3.get(), atol=atol, rtol=rtol)
            if verbose:
                print("duration (gpu) = {} s".format(t_gpu))
                print(
                    "duration (gpu, no transfer) = {} s".format(t_gpu_single2)
                )

        # Repeat with u, vr
        d1 = mri_exp_mult(A, u, vr)

        if config.have_cupy:
            vrg = cupy.asarray(vr)
            ug = cupy.asarray(u)
            d2 = mri_exp_mult(Ag, ug, vrg)
            cupy.testing.assert_allclose(d1, d2, atol=atol, rtol=rtol)
    return



"""Approximation of complex exponentials for off-resonance corrected MRI.

Notes
-----
Details of various fieldmap approximations are given in [1], [2].

References
----------
.. [1] J.A. Fessler, S. Lee, V.T. Olafsson, H.R. Shi, D.C. Noll. Toeplitz-based
    iterative image reconstruction for MRI with correction for magnetic field
    inhomogeneity. IEEE Trans. Signal Process. 2005; 53:3393–3402.

.. [2] B. P. Sutton, D. C. Noll, and J. A. Fessler. Fast, iterative image
    reconstruction for MRI in the presence of field inhomogeneities.
    IEEE Trans. Med. Imag. 2003; 22(2):178–188.

"""
import warnings

import numpy as np

from ._mri_exp_mult import mri_exp_mult
from mrrt.utils import (
    get_array_module,
    hist_equal,
    outer_sum,
    pinv_tol,
    profile,
)


__all__ = ["mri_exp_approx"]


@profile
def mri_exp_approx(
    ti=None,
    zmap=None,
    segments=6,
    approx_type="default",
    ctest=False,
    autocorr=False,
    verbose=False,
    tol=None,
    xp=None,
):
    """Approximation to exponentials for iterative MR image reconstruction.

    Generalizes "time segmentation" and "frequency segmentation" methods.
    This is used by the MRI_Operator object for field-corrected MR
    reconstruction.

    Parameters
    ----------
    ti : ndarray
        Sample times (in seconds).
    zmap : The field map to approximate.
        The real component corresponds to T2(*) decay while the imaginary
        component corresponds to off-resonance (in rad/s).
    segments : int, optional
        The number of segments to use for the approximation.
    approx_type : str, optional
        The type of field map approximation to apply. The default is
        histogram-based time segmentation.
        TODO: describe the options.

    Additional Parameters
    ---------------------
    ctest : bool, optional
        If True, a reduced size output with shape (L, nhist) instead of (L, N)
        is returned. This can be used for testing.
    autocorr : bool, optional
        If True, autocorrelate the fieldmap histogram (for the Toeplitz case).
    tol : float, "fro", or None, optional
        The tolerance used internally when computing the pseudo-inverse.
    xp : {numpy, cupy}
        The array module to use.



    Returns
    -------
    TODO

    Notes
    -----
    Details of various fieldmap approximations are given in [1], [2].

    References
    ----------
    .. [1] J.A. Fessler, S. Lee, V.T. Olafsson, H.R. Shi, D.C. Noll.
        Toeplitz-based iterative image reconstruction for MRI with correction
        for magnetic field inhomogeneity. IEEE Trans. Signal Process. 2005;
        53:3393–3402.
        :DOI:10.1109/TSP.2005.853152

    .. [2] B. P. Sutton, D. C. Noll, and J. A. Fessler. Fast, iterative image
        reconstruction for MRI in the presence of field inhomogeneities.
        IEEE Trans. Med. Imag. 2003; 22(2):178–188.
        :DOI:10.1109/TMI.2002.808360

    """
    """
    % in
    %    ti    [M,1]    sample times
    %    zmap    [N,1]    rate map: relax_map + 2i*pi*field_map
    %            *_map and ti should have reciprocal units.
    %            usually relax_map is 0, so zmap is purely imaginary!
    %    segments        number of components: 1 <= L << N
    %            Or, use {Linit, rmsmax}: give initial L to try,
    %            then increase L until RMS error <= rmsmax < 1.
    %            This is implemented only for 'hist,time,unif' type.
    %
    % options
    %    'ctest'    1|0    return C as [L,nhist] instead of full [L, N], for testing
    %    'autocorr'    1|0    autocorrelate the fmap histogram (for Toeplitz case)
    %    'tol'    tol    tolerance for pinv(), see pinv_tol() below.
    %            use {'fro'} for tol = 1e-8 * norm(X,'fro') (for large L)
    %    'approx_type'     what type of approximation (see choices below)
    %
    % out
    %    B    [M, L]    basis functions
    %    C    [L, N]    coefficients, such that B * C \approx exp(-ti * zmap.')
    %    hk,zk    [Ko, Kr]    histogram values and 'frequencies', if used
    %
    % type of approximation choices:
    %
    %    {'hist,time,unif', nhist}
    %        with uniform time samples and LS coefficients (using histogram).
    %        This approach works almost as well as the SVD method
    %        unless L is chosen so small that the error is large.
    %    'time,unif'
    %        Time segmentation with uniform time samples (LS coef).
    %        No histogram, so it is very slow.  not recommended.
    %
    %    {'hist,svd', nhist}
    %        Nearly optimal approach that uses an SVD and a zmap histogram.
    %        nhist is the # of histogram bins; about 40 is recommended.
    %
    %    {'hist,fs,unif', nhist}
    %        Not recommended since it works poorly except for uniform distn.
    %    {'hist,fs,prctile', nhist}
    %        Not recommended since it works quite poorly.
    %    {'hist,fs,lbg', nhist}
    %        Frequency segmentation methods (exponential bases)
    %        with different choices for the nominal frequency components.
    %        in all cases the coefficients are chosen by LS (Man, MRM, 1997).
    %
    %    for relaxation cases, nhist should be [Ko=#omap Kr=#rmap]
    %
    % Copyright 2004-7-1, Jeff Fessler, The University of Michigan
    """

    # defaults
    xp, on_gpu = get_array_module(zmap, xp)
    # make sure ti is on the same device as zmap
    ti = xp.asarray(ti)

    if verbose:
        from matplotlib import pylab as plt

    if approx_type == "default":
        if xp.any(zmap.real.ravel()):
            approx_type = ("hist,time,unif", [40, 10])
        else:
            approx_type = ("hist,time,unif", [40])

    atype = approx_type[0]

    segments = np.atleast_1d(segments)
    if len(segments) == 2:
        rmsmax = segments[1]
        segments = segments[0]
    elif len(segments) == 1:
        segments = segments[0]
        rmsmax = 1 / xp.finfo(float).eps
    else:
        raise ValueError(
            "segments must be an integer or a tuple containing an initial "
            "guess and an rms error tolerance"
        )
    zmap = zmap.ravel().reshape(-1, 1)  # [N,1]
    if ti.ndim > 1:
        warnings.warn("expected 1d ti")
    ti = ti.ravel()

    rmap = zmap.real
    fmap = zmap.imag / (2 * xp.pi)

    # histogram the field map
    warned = False
    hk = None
    zk = None
    if ("hist," in atype) or (ctest):
        nhist = approx_type[1]
        if len(nhist) == 1:  # fmap only

            if xp.any(rmap.ravel()):
                raise ValueError("rmap requires len(nhist) == 2")

            hk, zc = xp.histogram(zmap.imag, bins=nhist[0])
            if zc.dtype != zmap.imag.dtype:
                # use single precision if fieldmap was single precision
                zc = zc.astype(zmap.imag.dtype)

            # convert bin edges to bin centers to match matlab behavior!
            zc = zc[:-1] + xp.diff(zc) / 2.0

            # Note: skimage.exposure.histogram returns the bin centers instead
            # scipy.ndimage.histogram doesn't return bin edges
            # skimage.exposure.histogram(zmap, nhist[0]) -> seems to recenter
            # values

            zk = 0 + 1j * zc.ravel()
            if verbose:
                plt.figure(),
                bar_width = xp.mean(xp.diff(zc) / (2 * xp.pi)) * 0.66
                plt.bar(xp.asarray(zk.imag) / (2 * xp.pi), hk, width=bar_width)

            if autocorr:
                hk = xp.correlate(hk, hk, mode="full")
                zk = xp.arange(-(nhist[0] - 1), (nhist[0])) * (zk[1] - zk[0])
                if verbose:  # Todo
                    plt.figure(),
                    bar_width = xp.mean(xp.diff(zk.imag) / (2 * xp.pi)) * 0.8
                    plt.bar(
                        xp.asarray(zk.imag) / (2 * xp.pi), hk, width=bar_width
                    )
        else:
            hk, zc = hist_equal(
                xp.concatenate((zmap.imag, zmap.real), axis=1), nhist
            )
            if zc[0].dtype != zmap.imag.dtype:
                zc[0] = zc[0].astype(zmap.imag.dtype)
            if zc[1].dtype != zmap.real.dtype:
                zc[1] = zc[1].astype(zmap.real.dtype)
            zk = outer_sum(1j * zc[0], zc[1]).T  # [K1, K2]

            if autocorr:  # matlab version of this case by Valur Olafsson
                # TODO: haven't tested this case much
                import scipy.signal

                hk = scipy.signal.convolve2d(
                    hk, hk[::-1, ...], mode="full", boundary="fill", fillvalue=0
                )
                zc[0] = xp.arange(-(nhist[0] - 1), (nhist[0])) * (
                    zc[0][1] - zc[0][0]
                )
                zc[1] = xp.linspace(
                    2 * xp.min(zc[1]), 2 * xp.max(zc[1]), 2 * nhist[1] - 1
                )
                zk = outer_sum(1j * zc[0], zc[1]).T

        hk = hk.ravel()

        # GRL  Add this to conserve memory for cases where Eh is not needed
        if atype == "hist,svd" or verbose or (rmsmax < 1):
            Eh = xp.exp(
                xp.dot(-ti[:, np.newaxis], zk.ravel()[np.newaxis, :])
            )  # [N,K]

    #
    # SVD approach (optimal, to within histogram approximation)
    #
    if atype == "hist,svd":
        import scipy.linalg

        try:
            if xp is np:
                from scipy.sparse import spdiags
            else:
                from cupyx.scipy.sparse import spdiags
            # weighted signals
            Ew = xp.asarray(
                xp.asmatrix(Eh)
                * spdiags(xp.sqrt(hk), 0, len(hk), len(hk), format="csr")
            )
        except ImportError:
            Ew = xp.dot(Eh, xp.diag(xp.sqrt(hk)))
        if xp is np:
            U, s, V = scipy.linalg.svd(Ew, full_matrices=False)
            B = U[:, :segments]  # [M, L] keep desired components
        else:
            U, s, V = scipy.linalg.svd(Ew.get(), full_matrices=False)
            B = U[:, :segments]  # [M, L] keep desired components
            B = xp.asarray(B)

    #
    # time segmentation approaches (recommended)
    #
    elif (atype == "time,unif") or (atype == "hist,time,unif"):

        rms = xp.inf
        while (rms > rmsmax) and (segments < 40):
            # time sample locations [0 ... end]
            if segments == 1:
                tl = ti.mean()
            else:
                p = [100 * q / (segments - 1) for q in range(segments)]
                tl = xp.asarray(xp.percentile(ti, q=p))
                if tl.dtype != zmap.imag.dtype:
                    tl = tl.astype(zmap.imag.dtype)

            if (rmsmax < 1) or ctest or (atype == "hist,time,unif"):
                # Ch is shape (L, K)
                # Ch = xp.exp(xp.dot(-tl[:, np.newaxis], zk[np.newaxis, :]))

                Ch = xp.exp(
                    xp.dot(-tl[:, np.newaxis], zk.ravel()[np.newaxis, :])
                )

            if (segments > 9) and (tol is None) and (not warned):
                warnings.warn("warning: For large L, trying tol='fro'")
                warned = True
                if isinstance(tol, list):
                    tol[0] = "fro"
                else:
                    tol = "fro"

            if atype == "time,unif":
                # [L, N] - classic TS
                C = xp.exp(xp.dot(-tl[:, np.newaxis], zmap.reshape(1, -1)))
                X = C.T  # [N,L]
                X = pinv_tol(X, tol).T  # [N,L]
                B = mri_exp_mult(
                    X.conj(),
                    zmap.ravel(),
                    # zmap.astype(xp.promote_types(zmap.dtype, xp.float64)),
                    ti,
                ).T
            elif atype == "hist,time,unif":  # Fessler section 5.4.3.5
                if xp is np:
                    from scipy.sparse import spdiags
                else:
                    from cupyx.scipy.sparse import spdiags
                W = spdiags(xp.sqrt(hk), 0, len(hk), len(hk), format="csr")
                if W.dtype != zmap.imag.dtype:
                    # use single precision if fieldmap was single precision
                    W = W.astype(zmap.imag.dtype)
                if xp is np:
                    P = pinv_tol(W * Ch.T, tol) * W  # [L,K], weighted pinv
                else:
                    # TODO: fix pinv_tol on the GPU.
                    #       for now run it on the CPU and then copy result back to the GPU
                    import cupy

                    P = cupy.asarray(
                        pinv_tol(W.get() * Ch.T.get(), tol) * W.get()
                    )  # [L,K], weighted pinv
                # print("zk.dtype = {}".format(zk.dtype))
                # print("P.dtype = {}".format(P.dtype))

                if False and verbose:
                    # rP = numpyutils.matrixrank(P);
                    rP = xp.linalg.matrix_rank(P)  # crashes?

                    if rP != segments:
                        warnings.warn("rank=%d < L=%d" % (rP, segments))
                B = mri_exp_mult(P.conj().T, zk.ravel(), ti).T

            if rmsmax < 1:
                segments += 1
                Ep = xp.dot(B, Ch)
                atmp = xp.abs(Eh - Ep)
                atmp *= atmp
                mse = xp.mean(atmp, axis=0)
                rms = xp.dot(xp.sqrt(mse), hk) / xp.sum(hk)  # hk-weighted rms
                print("rms={} at segments={}".format(rms, segments))
                if verbose:  # GRL added
                    from pyvolplot import subplot_stack

                    iworst = xp.argmax(mse)
                    ik = [
                        0,
                        np.floor(nhist[0] / 4),
                        2 * np.floor(nhist[0] / 4),
                        3 * np.floor(nhist[0] / 4),
                        nhist[0] - 1,
                        np.prod(nhist) - 1,
                        iworst,
                    ]
                    fig = subplot_stack(1000 * ti, Eh[:, ik], colors=["g", "k"])
                    fig = subplot_stack(
                        1000 * ti,
                        Ep[:, ik],
                        colors=["b--", "r--"],
                        fig=fig,
                        title="True and Approx, segments={}".format(segments),
                    )

            else:
                break  # check

        if segments == 40:
            raise ValueError("max segments reached!?")

        if ctest:
            C = Ch
        elif not atype == "time,unif":
            C = xp.dot(-tl[:, np.newaxis], zmap.reshape(1, -1))  # [L, N]
            xp.exp(C, out=C)

        return B, C, hk, zk

    #
    # freq. segmentation approaches
    #
    elif atype[:8] == "hist,fs,":

        if atype == "hist,fs,unif":  # uniform spacing
            fl = xp.linspace(xp.min(fmap), xp.max(fmap), segments + 2)
            fl = fl[1:-1]
            rl = 0  # lazy: 'uniform' in 2D seems too arbitrary
            # this may stink if rmap is nonzero.

        elif atype == "hist,fs,prctile":  # histogram percentiles
            p = [100 * q / (segments + 1) for q in range(1, segments + 1)]
            fl = xp.array(xp.percentile(fmap, q=p))
            rl = 0  # lazy again

        elif atype == "hist,fs,kmeans":
            # requires scikit-learn package
            try:
                from sklearn.cluster import KMeans
            except ImportError as e:
                print("hist,fs,kmeans requires sklearn")
                raise (e)
            k_means = KMeans(n_clusters=int(segments), n_init=10, n_jobs=-1)
            if xp is not np:
                fmap_np = fmap.get()
                rmap_np = rmap.get()
            else:
                fmap_np = fmap
                rmap_np = rmap

            if xp.any(rmap):
                k_means.fit(
                    xp.stack((rmap_np.ravel(), fmap_np.ravel()), axis=-1)
                )
                centers = k_means.cluster_centers_
                si = xp.argsort(centers[:, 1])
                fl = xp.sort(centers[:, 1][si])
                rl = xp.sort(centers[:, 0][si])
            else:
                k_means.fit(fmap.reshape((-1, 1)))
                centers = k_means.cluster_centers_.squeeze()
                fl = xp.sort(xp.asarray(centers))
                rl = 0

        elif atype == "hist,fs,lbg":  # LBG quantization of histogram
            raise ValueError(
                "Haven't implemented the Linde–Buzo–Gray (LBG) algorithm in "
                "python.  Try hist,fs,kmeans instead"
            )
        else:
            raise ValueError("fs type {} unknown".format(atype))

        zl = rl + (2j * xp.pi) * fl
        B = xp.exp(
            xp.dot(-ti[:, np.newaxis], zl.ravel()[np.newaxis, :])
        )  # (M, L)

    else:
        raise ValueError("type {} unknown".format(atype))

    # given the basis, now compute the LS coefficients,
    # i.e., (B'*B) \ B' * E
    Bpp = pinv_tol(B, tol).T.conj()  # [M, L]

    if ctest:
        # [L, K] coefficients
        C = mri_exp_mult(Bpp, ti, zk.ravel())
    else:
        # [L, N] coefficients
        C = mri_exp_mult(Bpp, ti, zmap.ravel())

    # make sure outputs are all arrays
    B = xp.asarray(B)
    C = xp.asarray(C)
    if hk is not None:
        hk = xp.asarray(hk)
        zk = xp.asarray(zk)
    return B, C, hk, zk


if False:
    from mrrt.mri.sim import generate_fieldmap

    zmap = generate_fieldmap((64, 64), 80)



#!/usr/bin/env python
import os
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

PACKAGES = find_packages()

# Get version and release info, which is all stored in mrrt/mri/version.py
ver_file = os.path.join("mrrt", "mri", "version.py")
with open(ver_file) as f:
    exec(f.read())
# Give setuptools a hint to complain if it's too old a version
# 24.2.0 added the python_requires option
# Should match pyproject.toml
SETUP_REQUIRES = ["setuptools >= 24.2.0"]
# This enables setuptools to install wheel on-the-fly
SETUP_REQUIRES += ["wheel"] if "bdist_wheel" in sys.argv else []


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


cmdclass = {"test": PyTest}

opts = dict(
    name=NAME,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url=URL,
    download_url=DOWNLOAD_URL,
    license=LICENSE,
    classifiers=CLASSIFIERS,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    platforms=PLATFORMS,
    version=VERSION,
    packages=PACKAGES,
    package_data=PACKAGE_DATA,
    install_requires=REQUIRES,
    python_requires=PYTHON_REQUIRES,
    setup_requires=SETUP_REQUIRES,
    requires=REQUIRES,
    cmdclass=cmdclass,
)

if __name__ == "__main__":
    setup(**opts)



"""
example script to use in testing for CUDA memory leaks via:

cuda-memcheck python -m pycuda.debug example_debug.py

To check thread race conditions:
cuda-memcheck --tool racecheck python example_debug.py

To check for uninitialized global memory accesses
cuda-memcheck --tool initcheck python example_debug.py

To check for incorrect synchronization
cuda-memcheck --tool synccheck python example_debug.py


Can also profile using nvidia's nvprof

nvprof --print-gpu-trace python example_debug.py

"""

from mrrt.operators.bench.bench_mri import (
    bench_mri_2d_nocoils_nofieldmap,
    bench_mri_2d_16coils_fieldmap,
)

bench_mri_2d_nocoils_nofieldmap(
    navg_time=1, include_GPU=True, include_CPU=False, print_timings=True
)

bench_mri_2d_16coils_fieldmap(
    navg_time=1, include_GPU=True, include_CPU=False, print_timings=True
)
